# -*- coding: utf-8 -*-
"""
Classes for the handling the fields.
"""
import collections
import inspect
import logging
import typing as T

try:
    # python >= 3.7.2
    from typing import OrderedDict
except ImportError:
    # python >= 3.6
    from typing import MutableMapping as OrderedDict

import warnings

import numpy as np
from scipy.ndimage import find_objects, label, binary_dilation
import shapely.geometry
import xarray as xr
from xarray.core.utils import either_dict_or_kwargs
from xarray.core.dataarray import _LocIndexer

import cerbere.cfconvention as cf

T_Coverage = T.Literal[
    'image', 'thematicClassification', 'physicalMeasurement',
    'auxiliaryInformation', 'qualityInformation', 'referenceInformation',
    'modelResult', 'coordinate'
]


FIELD_ATTRS = [
    'standard_name',
    'authority',
    'units',
    'valid_min',
    'valid_max',
    '_FillValue'
]
FIELD_EXCL_ATTRS = [
    'long_name'
]


def _has_nonfinite(x):
    if np.issubdtype(x.dtype, np.dtype(np.datetime64)):
        return np.isnat(np.asarray(x).sum())
    elif (np.issubdtype(x.dtype, np.dtype('O'))
              or np.issubdtype(x.dtype, np.dtype('U'))):
        return True
    return np.isnan(np.asarray(x).sum())


class _CLocIndexer(_LocIndexer):

    def __getitem__(self, key) -> xr.DataArray:
        return super().__getitem__(key)

    def __setitem__(self, key, value) -> None:
        if _has_nonfinite(value):
            if self.data_array.cb.fill_value is None:
                # masked value are set on a data array which had no fill value
                # xarray will create a default fill value but without filling
                # in the encoding attribute; we fix that.
                self.data_array.cb._cfarray.encoding['_FillValue'] = \
                    cf.default_fill_value(
                        self.data_array.cb.science_dtype)

        self.data_array.loc[key] = value


@xr.register_dataarray_accessor("cb")
class DataArrayAccessor:
    """Describes a scientific data array. It contains data and
    metadata attributes.

    This is an extension of xarray's DataArray with stricter requirements on
    attributes and the encoding of masked arrays.
    """
    def __init__(self, xarray_obj):
        self._orig_array = None

        self._cfarray = xarray_obj
        self.is_cerberized: bool = False

        self.longitude180: bool = True

        self.cerberize()

    @property
    def cfarray(self) -> xr.DataArray:
        """Returns a CF normalized version of the dataset."""
        if self._cfarray is None:
            # default standardization - may not be enough for some products
            # that have more unusual none-conformity
            self.cerberize()

        return self._cfarray

    def cerberize(self):
        """
        Do whatever is necessary to make the variable compliant to Cerbere
        conventions (based on CF, ACDD).

        May not be enough for some products that have more unusual
        none-conformity.
        """
        if self.is_cerberized:
            # already cerberized
            return

        # back-up original version for encoding information - NOTE: the
        # modified array has to be the array of of the accessor or strange
        # behaviour may occur (like with loc function).
        # @TODO may have a memory imprint if some data copy occur - may be
        #  better to only save variable and global encoding
        self._orig_array = self._cfarray.copy(deep=False)

        # guess science data type
        self._cfarray.encoding['science_dtype'] = self._guess_science_dtype()

        # fix some possible mismatch between the science data dtype and the
        # fill value dtype
        self._fix_fill_value()

        self.is_cerberized = True

    def _fix_fill_value(self) -> None:
        """Fix the data fill value if its dtype is mismatching the
        science data dtype.

        Cases where such mismatch may occur:

        * data are read from a file which uses scaling (through `scale_factor`
          and `add_offset` CF attributes). The _FillValue is given for the
          storage type (which may be an integer even if the data are unscaled to
          a float type.
        * xarray creates automatically some _FillValue in its encoding
          attribute (for instance when setting np.nan in the data): for
          floating values, the created _FillValue may be a float64 instead of a
          float32 when float32 is the actual data dtype.
        """
        # enforce fill value for datetime64 and timedelta64 arrays
        if np.issubdtype(self.cfarray.encoding['science_dtype'],
                         np.datetime64):
            self.cfarray.encoding['_FillValue'] = np.datetime64('NaT')
            return
        if np.issubdtype(self.cfarray.encoding['science_dtype'],
                         np.timedelta64):
            self.cfarray.encoding['_FillValue'] = np.timedelta64('NaT')
            return

        # @ TODO : check before uncommenting -> conflict with following code
        # if ('_FillValue' not in self.cfarray.encoding and '_FillValue' not in
        #         self.cfarray.attrs):
        #     return

        # xarray sometimes holds _FillValue in attrs...
        fv = self.cfarray.encoding.get(
            '_FillValue', self.cfarray.attrs.get('_FillValue'))

        if fv is None and _has_nonfinite(self.cfarray):
            # if the array has none-finite values, it must have a fill value
            # different from None => affect one
            fv = cf.default_fill_value(self.cfarray.encoding['science_dtype'])
            self.cfarray.encoding['_FillValue'] = fv

        if fv is None:
            return

        data_dtype = self.cfarray.encoding['science_dtype']

        # data_dtype = self.cfarray.dtype
        fv_dtype = np.dtype(type(fv))

        if data_dtype != fv_dtype:
            # force the _FillValue to same as science data dtype
            self.cfarray.encoding['_FillValue'] = data_dtype.type(
                self.cfarray.encoding['_FillValue'])

        # if np.issubdtype(data_dtype, np.floating) and \
        #         np.issubdtype(fv_dtype, np.floating):
        #     # fix the _FillValue
        #     self.cfarray.encoding['_FillValue'] = cf.default_fill_value(
        #         data_dtype)


        # fv_dtype = None
        # if self.cfarray.encoding.get('_FillValue') is not None:
        #     fv_dtype = np.dtype(type(self.cfarray.encoding.get('_FillValue')))
        # if len(encoding_dtype) == 0 and fv_dtype is not None:
        #     # no data scaling: verify the _FillValue is consistent with the
        #     # data dtype as xarray creates automatically some _FillValue in
        #     # encoding attribute (for instance when setting np.nan in the data)
        #     # for floating values, the created _FillValue may be a float64
        #     # instead of a float32
        #     if np.issubdtype(self.cfarray.dtype, np.floating) and \
        #             np.issubdtype(fv_dtype, np.floating):
        #         # fix the _FillValue
        #         self.cfarray.encoding['_FillValue'] = cf.default_fill_value(
        #             self.cfarray.dtype)
        #         fv_dtype = self.cfarray.dtype
        #     encoding_dtype = {fv_dtype}
        # elif fv_dtype is not None:
        #     encoding_dtype = encoding_dtype.union({fv_dtype})

    def _guess_science_dtype(self) -> np.dtype:
        """
        Guess the scientific dtype from data read on file.

        xarray sometimes transform a DataArray dtype, for instance when
        converting fill values from a numpy MaskedArray to NaN which are only
        defined for floating types. Integer, bytes, etc... array can then be
        converted to float. This happens for instance when reading masked
        integer variables from a NetCDF file.

        In cerbere we keep track of the "true" scientific dtype to return for
        instance numpy ndarray or MaskedArray objects in their true
        original dtype.
        """
        if 'science_dtype' in self._cfarray.encoding:
            return self._cfarray.encoding['science_dtype']

        if np.issubdtype(self._cfarray.dtype, np.datetime64):
            # ignore internal dtype encoding
            return self._cfarray.dtype

        # if data are read from a file, the scaling CF attributes,
        # if present, give the scientific data dtype
        encoding_dtype = {np.dtype(type(_)) for _ in [
            self._cfarray.encoding.get('scale_factor'),
            self._cfarray.encoding.get('add_offset')]
            if _ is not None
        }

        if len(encoding_dtype) == 0:
            # no encoding : assume the array dtype is the same as the
            # scientific dtype
            return self._cfarray.dtype

        if len(encoding_dtype) == 1:
            # all encoding scaling attributes are consistent
            return next(iter(encoding_dtype))

        if len(encoding_dtype) > 1:
            dtypes = {
                _: np.dtype(type(self._cfarray.encoding[_]))
                for _ in ['scale_factor', 'add_offset']
                if _ in self._cfarray.encoding
            }
            logging.warning(
                f'inconsistent dtypes in CF encoding attributes: {dtypes}. '
                f'This should be fixed by the producer of these data.')

            # assuming the scale factor is more correct
            return dtypes['scale_factor']

    @property
    def fill_value(self) -> T.Union[T.Any, None]:
        """The value for missing data in the field"""
        return self.cfarray.encoding.get('_FillValue')

    @fill_value.setter
    def fill_value(self, fill_value: T.Any):
        """set the value for missing data"""
        if fill_value is None:
            # @TODO : raise exception if any masked value ?
            self.cfarray.encoding['_FillValue'] = None
            return

        # do some checks on consistency with data type
        if self.science_dtype is not None and not np.issubdtype(
                np.dtype(type(fill_value)), self.science_dtype):
            raise TypeError(
                f'the dtype {np.dtype(type(fill_value))} for the fill_value '
                f'is not consistent with data dtype {self.science_dtype}')

        if self.cfarray.cb.fill_value is None:
            self.cfarray.encoding['_FillValue'] = fill_value
            return

        if self.cfarray.cb.fill_value != fill_value:
            # a fill value is already set and different from the new one,
            # the existing masked values need to be replaced
            logging.warning(
                f'Changing the fill values in array {self.cfarray.name} from '
                f'{self.cfarray.cb.fill_value} to {fill_value}. This '
                f'should be avoided for memory and performance issues.')
            self._cfarray = self._replace_fill_value(
                self.cfarray, self.cfarray.cb.fill_value, fill_value)

        self.cfarray.encoding['_FillValue'] = fill_value

    @property
    def precision(self) -> T.Union[int, None]:
        """The least significant digit for the values of the field.

        Used for storage optimization when writing on disk. The least
        significant digit is the power of ten of the smallest decimal place in
        the data that is a reliable value.
        """
        return self.cfarray.encoding.get('least_significant_digit')

    @precision.setter
    def precision(self, precision: int):
        """Set the least significant digit for the values of the field.

        Used for storage optimization when writing on disk. The least
        significant digit is the power of ten of the smallest decimal place in
        the data that is a reliable value.
        """
        self.cfarray.encoding['least_significant_digit'] = precision

    @property
    def valid_min(self) -> T.Union[T.Any, None]:
        """The minimum valid value in the field data"""
        return self.cfarray.attrs.get('valid_min')

    @valid_min.setter
    def valid_min(self, value: T.Any):
        """set the minimum valid value (``valid_min`` CF attribute)"""
        self.cfarray.attrs['valid_min'] = value

    @property
    def valid_max(self) -> T.Union[T.Any, None]:
        """The maximum valid value in the data (``valid_max`` CF attribute)"""
        return self.cfarray.attrs.get('valid_max')

    @valid_max.setter
    def valid_max(self, value: T.Any):
        """set the maximum valid value (``valid_max`` CF attribute)"""
        self.cfarray.attrs['valid_max'] = value

    @property
    def scale_factor(self) -> T.Union[T.Any, None]:
        """The encoding scaling factor for the data (``scale_factor`` CF
        attribute)"""
        return self.cfarray.encoding.get('scale_factor')

    @scale_factor.setter
    def scale_factor(self, value: T.Any):
        """set the encoding scaling factor (``scale_factor`` CF attribute)"""
        self.cfarray.encoding['scale_factor'] = value

    @property
    def add_offset(self) -> T.Union[T.Any, None]:
        """The encoding offset factor for the data (``add_offset`` CF
        attribute)"""
        return self.cfarray.encoding.get('add_offset')

    @add_offset.setter
    def add_offset(self, value: T.Any):
        """set the encoding offset factor (``scale_factor`` CF attribute)"""
        self.cfarray.encoding['add_offset'] = value

    @property
    def axis(self) -> T.Optional[str]:
        """The spatiotemporal axis represented by this variable (
        ``axis`` CF attribute). None if is not set as an axis."""
        return self.cfarray.attrs.get('axis')

    @property
    def units(self) -> T.Union[str, None]:
        """The variable data units (``units`` CF attribute)"""
        return self.cfarray.encoding.get(
            'units', self.cfarray.attrs.get('units'))

    @units.setter
    def units(self, units: str):
        """Set the variable data units (``units`` CF attribute)"""
        if 'units' in self.cfarray.encoding:
            self.cfarray.encoding['units'] = units
        else:
            self.cfarray.attrs['units'] = units

    @property
    def long_name(self) -> T.Union[str, None]:
        """The variable description (``long_name`` CF attribute)"""
        return self.cfarray.attrs.get('long_name')

    @long_name.setter
    def long_name(self, long_name: str) -> None:
        """set the variable description (``long_name`` CF attribute)"""
        self.cfarray.attrs['long_name'] = long_name

    @property
    def standard_name(self) -> T.Union[str, None]:
        """The variable standard name (``standard_name`` CF attribute)"""
        return self.cfarray.attrs.get('standard_name')

    @standard_name.setter
    def standard_name(self, standard_name: str) -> None:
        """set the standard_name (``standard_name`` CF attribute)"""
        self.cfarray.attrs['standard_name'] = standard_name

    @property
    def ancillary_variables(self):
        """CF Attribute: identifies a variable that contains closely
        associated data, e.g., the measurement uncertainties of instrument data.
        """
        return self.cfarray.attrs['ancillary_variables']

    @ancillary_variables.setter
    def ancillary_variables(self, value: T.List[str]):
        self.cfarray.attrs['ancillary_variables'] = value

    @property
    def coverage_content_type(self) -> T.Union[T_Coverage, None]:
        """ACDD 1.3 attribute; An ISO 19115-1 code to indicate the source of the
        data (image, thematicClassification, physicalMeasurement,
        auxiliaryInformation, qualityInformation, referenceInformation,
        modelResult, or coordinate).
        """
        return self.cfarray.attrs['coverage_content_type']

    @coverage_content_type.setter
    def coverage_content_type(self, value: T_Coverage):
        if value not in cf.COVERAGE_CONTENT_TYPE:
            raise ValueError(
                f'value {value} is not allowed. It should be one of the '
                f'following: {list(cf.COVERAGE_CONTENT_TYPE)}')
        self.cfarray.attrs['coverage_content_type'] = value

    @property
    def science_dtype(self) -> np.dtype:
        """The scientific dtype of the data (ignores xarray
        internal conversion)"""
        if self.cfarray.encoding.get('science_dtype') is None:
            self.cfarray.encoding['science_dtype'] = self._guess_science_dtype()
        return self.cfarray.encoding['science_dtype']

    @science_dtype.setter
    def science_dtype(self, value: np.dtype):
        if not isinstance(np.dtype(value), np.dtype):
            raise TypeError(f'{value} is not a numpy dtype')
        self.cfarray.encoding['science_dtype'] = np.dtype(value)

    def __str__(self):
        result = "\nField : '%s'\n" % self.cfarray.name
        if ('long_name' in self.cfarray.attrs and
                self.cfarray.attrs['long_name'] is not None):
            result += '    {}\n'.format(self.cfarray.attrs['long_name'])
        result += '\n'

        # dimensions
        result = result + '    dimensions :\n'
        dims = collections.OrderedDict(self.cfarray.sizes.items())
        for dim, size in dims.items():
            result += '      # {} : {}\n'.format(dim, size)

        attrs = self.cfarray.attrs.items()

        # standard attributes
        result = result + '    standard CF attributes :\n'
        for att, val in attrs:
            if att in FIELD_ATTRS:
                result += '      # {} : {}\n'.format(att, val)
        result += '      # fill_value : {}\n'.format(str(self.fill_value))

        # free form attributes
        result = result + '    other attributes :\n'
        for att, val in attrs:
            if att not in FIELD_ATTRS and att not in FIELD_EXCL_ATTRS:
                result = result + '      # {} : {}\n'.format(att, val)

        return result

    def is_axis(self, axis: str) -> bool:
        """Returns true if the field is the coordinate field for the given axis.

        Args:
            axis: axis name (among X, Y, Z, T)
        """

    def _broadcast_coords(
            self,
            values: xr.DataArray,
            broadcast_coords: T.List[xr.DataArray]):
        """
        Broadcast the values along a list of existing and new coordinates.

        Existing coordinates are ignored, except for ordering the dimensions
        in the output array (following the order given in broadcast_coords).
        If existing coordinates of the array are not listed in
        broadcast_coords, they follow the ones given in broadcast_coords in
        the order of the dimensions of the output array.

        Args:
            coords: coordinates along which to broadcast

        Returns:

        """
        # brodcast values over the requested dimensions
        cdims = [_.name for _ in broadcast_coords]
        for coord in broadcast_coords[::-1]:
            dim = coord.name
            if dim in self._cfarray.dims:
                # this coordinate is already a coordinate of the current array:
                # no broadcasting needed
                continue

            values = values.broadcast_like(coord)

        # # @TODO check more cases
        # squeezed_dims = [
        #     # _ for _ in self.dims
        #     _ for _ in self._cfarray.dims
        #     if _ in self.geocoordnames and _ not in broadcast_coords
        # ]
        # values = values.squeeze(dim=squeezed_dims, drop=True)
        # rearranged_dims = cdims + [
        #     _ for _ in values.dims if _ not in cdims
        # ]
        #
        # re-order dims
        # if len(values.dims) > 0 & (values.dims != rearranged_dims):
        #     values = values.transpose(
        #         *(list(rearranged_dims)), transpose_coords=True)
        rearranged_dims = cdims + [
             _ for _ in values.dims if _ not in cdims
        ]
        if len(values.dims) > 0 & (values.dims != rearranged_dims):
            values = values.transpose(
                *(list(rearranged_dims)), transpose_coords=True)

        return values

    def broadcast(self,
                  target: T.Union[
                      xr.DataArray,
                      T.List[T.Union[xr.DataArray, str]]],
                  feature = None):
        """Broadcast the values of a DataArray along the coordinates of
        another DataArray or a list of coordinates.
        """
        if isinstance(target, xr.DataArray):
            return self._cfarray.broadcast_like(target)
        elif isinstance(target, list):
            if len(target) == 0:
                raise ValueError('Cannot broadcast along an empty list of '
                                 'coordinates')
            if isinstance(target[0], xr.DataArray):
                return self._broadcast_coords(self._cfarray, target)
            elif isinstance(target[0], str):
                if feature is None:
                    raise ValueError(
                        'feature object must be provided when broadcasting by '
                        'coordinate names')
                return self._broadcast_coords(
                    self._cfarray, [feature.ds[_] for _ in target])

        raise TypeError('target type not accepted for broadcast operation')

    def isel(
            self,
            indexers = None,
            padding: bool = False,
            broadcast_coords: T.List[xr.DataArray] = None,
            as_science_dtype: bool = False,
            as_masked_array: bool = False,
            **kwargs: T.Any):
        """

        Args:
            padding: pad the result with fill values where slices are out of the
                field dimension limits. Default is False.
            as_science_dtype: returns the data in their science dtype,
                replacing NaN with fill values
            **kwargs:

        Returns:

        """
        # separate indexer kargs from isel other kwargs
        # remove dims of indexer not in the variable
        isel_signature = inspect.getfullargspec(self.cfarray.isel).args
        isel_kwargs = {k: v for k, v in kwargs.items() if k in isel_signature}
        indx_kwargs = {k: v for k, v in kwargs.items() if k not in
            isel_signature}
        sel = either_dict_or_kwargs(indexers, indx_kwargs, "isel")

        if padding:
            sel = {
                dim: slice(max(0, sl.start), sl.stop)
                for dim, sl in sel.items()}

        # remove from indexer dims not in the DataArray object
        reduced_sel = {k: v for k, v in sel.items() if k in self.cfarray.dims}

        # select
        if len(reduced_sel) == 0:
            arr = self.cfarray
        else:
            arr = self._cfarray.isel(**reduced_sel, **isel_kwargs)

        # broadcast to extra dims if requested
        if broadcast_coords is not None:

            arr = self._broadcast_coords(
                arr,
                broadcast_coords=[
                    coord.isel({k: v for k, v in sel.items() if k in
                                coord.dims})
                    for coord in broadcast_coords])

        if as_science_dtype:
            arr = arr.fillna(self.fill_value).astype(
                self.science_dtype, copy=False)

        if padding:
            arr = self._pad_data(
                self.cfarray, arr,
                either_dict_or_kwargs(indexers, indx_kwargs, "isel"))

        # fix longitudes in [0, 360] range instead of [-180, 180]
        if arr.name == 'lon' and self.longitude180:
            if arr.max() >= 180.:
                if isinstance(arr, np.ndarray) and not arr.flags.writeable:
                    arr = arr.copy()
                arr[arr >= 180.] -= 360

        if as_masked_array:
            arr = arr.cb.to_masked_array()

        return arr

    @property
    def loc(self) -> _CLocIndexer:
        """Attribute for location based indexing like pandas."""
        return _CLocIndexer(self._cfarray)

    @property
    def mask(self) -> T.Union[np.ndarray, bool]:
        """Return the mask of the data (like in numpy MaskedArray)

        Masked values are NaN and values equal to the fill value
        """
        # mask NaN
        mask = np.ma.make_mask(
            ~np.isfinite(self.cfarray), shrink=True, copy=False)

        # mask where fill value
        if self.fill_value is not None:
            fv_mask = np.ma.make_mask(
                self.cfarray == self.fill_value, shrink=True,
                copy=False)

            mask = np.ma.mask_or(mask, fv_mask)

        return mask

    @classmethod
    def _replace_fill_value(cls, arr, previous, new):
        if previous not in arr:
            return arr

        encoding = arr.encoding.copy()
        attrs = arr.attrs.copy()
        # @TODO does a copy? should replace inline
        arr = arr.where(arr != previous, new)

        # restore attrs (mimic option keep_attrs in where method missing
        # in older versions of xarray)
        arr.encoding = encoding
        arr.attrs = attrs

        return arr

    def to_masked_array(self) -> np.ma.MaskedArray:
        """

        Returns:

        """
        arr = self.cfarray.to_masked_array(copy=False)

        if self.science_dtype != self.cfarray.dtype:
            # convert to original science dtype
            arr = arr.astype(self.science_dtype, copy=False)

        if self.fill_value is not None:
            if arr.fill_value is not None and arr.fill_value != self.fill_value:
                # xarray has defined its own fill value and cerbere has its
                # own - we can have only one fill value in the data
                arr = np.ma.masked_equal(
                    self._replace_fill_value(
                        arr, arr.fill_value, self.fill_value),
                    self.fill_value, copy=False)

            arr.set_fill_value(self.fill_value)

        return arr

    @classmethod
    def _pad_data(
            cls,
            array: xr.DataArray,
            subset: xr.DataArray,
            index: T.Optional[T.Mapping[str, slice]]
    ) -> xr.DataArray:
        """
        pad with fill values the ``subset`` array extracted from ``array``
        where ``index`` is beyond the limits of ``array``.
        """
        if len(array.shape) == 0:
            # dimensionless field
            return subset

        # determine required padding for each dim
        pad_edges = []
        for dim in list(array.dims):
            if dim in index:
                dslice = index[dim]
                pad_edges.append([
                    abs(min(0, dslice.start)),
                    abs(min(0, array.sizes[dim] - dslice.stop))
                ])
            else:
                pad_edges.append([0, 0])

        if all([_ == [0, 0] for _ in pad_edges]):
            # no padding necessary
            return subset

        # pad data
        sdtype = subset.cb.science_dtype
        if not np.issubdtype(subset.dtype, np.floating):
            # transform to float64 as xarray would do
            subset = subset.astype(np.float64, copy=False)

        padded_values = np.pad(
            subset.to_masked_array(copy=False), pad_edges, 'constant',
            constant_values=np.nan)
        dims = {dim: padded_values.shape[i] for i, dim in enumerate(array.dims)}

        # pad coordinates
        coords = {}
        for k_coord, v_coord in subset.coords.items():

            # @TODO check side effects first
            # if index is None or \
            #         len(set(index.keys()).intersection(set(v_coord.dims))) == 0:
            #     coords[k_coord] = v_coord
            #     continue

            c_pad_edges = tuple([
                tuple(pad_edges[i]) for i, dim in enumerate(array.dims)
                if dim in v_coord.dims
            ])
            c_fill_value = subset[k_coord].cb.fill_value
            if c_fill_value is None:
                # create a valid fill value if there was none so far
                c_fill_value = cf.default_fill_value(
                    subset[k_coord].cb.science_dtype)
                v_coord.cb.fill_value = c_fill_value

            padded_arr = np.ma.masked_values(np.pad(
                v_coord.to_masked_array(copy=False),
                pad_width=c_pad_edges,
                mode='constant',
                constant_values=c_fill_value), c_fill_value, copy=False)

            coords[k_coord] = xr.DataArray(
                padded_arr,
                dims=v_coord.dims,
                attrs=v_coord.attrs
            )
            coords[k_coord].encoding = v_coord.encoding

        padded_values = xr.DataArray(
            data=padded_values,
            dims=dims,
            coords=coords,
            attrs=subset.attrs)
        padded_values.encoding = subset.encoding
        padded_values.cb.science_dtype = sdtype

        return padded_values

    def bitmask_or(self, meanings):
        """helper function to test specific bits in a CF compliant mask

        Bit (or flag) fields are arrays of integers where each bit has a
        specific meaning, described in a ``flag_meaning`` field attribute.
        Providing a list of the meanings to be tested in ``meaning``, a boolean
        mask if built, using the `or` logical operator, with a True value
        everywhere at least one of the provided meanings is set.

        The field must defined as in CF convention, with ``flags_masks``
        and ``flag_meanings`` field attributes.

        Args:
            meanings(list or str): a list of the meanings that have to be set
                or a str if only one bit is tested. Available meanings are
                listed in the ``flag_meanings`` attribute of the field.

        Returns:
            A boolean array.
        """
        if ('flag_meanings' not in self.cfarray.attrs or
                'flag_masks' not in self.cfarray.attrs):
            raise ValueError(
                'This is not mask field. Either flag_meanings or flag_masks '
                'is missing.'
            )

        # transform mask attributes to list if it is not the case
        allmeanings = self.cfarray.attrs['flag_meanings']
        if isinstance(allmeanings, str):
            allmeanings = [_ for _ in allmeanings.split(' ') if _ != '']

        allmasks = self.cfarray.attrs['flag_masks']
        if isinstance(allmasks, str):
            allmasks = [_ for _ in allmasks.split(' ') if _ != '']

        criteria = meanings
        if isinstance(meanings, str):
            criteria = [meanings]

        # calculate sum (and) of all mask bits requested to be set
        masksum = 0
        for criterium in criteria:
            try:
                bit = allmeanings.index(criterium)
            except:
                raise ValueError('Unknown flag meaning {}'.format(criterium))
            masksum += allmasks[bit]

        # calculate mask
        return self.cfarray & int(masksum) != 0

    def clip(
            self,
            geometry: shapely.geometry.Polygon,
            masked: bool = False,
            continuity_threshold: int = 0,
            as_masked_array: bool = True,
            **kwargs) -> T.List[T.Union[xr.DataArray, np.ma.masked_array]]:
        """Return the data of a field within a geographical area.

        Extract the subset(s) of the Dataset including the whole area geometry,
        defined as a polygon of lon/lat coordinates.

        Any keywords from :meth:`get_values()` can be applied here too.

        Args:
            geometry (shapely.geometry.Polygon): boundaries of the clipped data,
                as a polygon of lon, lat coordinates.

            continuity_threshold (int): maximum number of non-contiguous
                measurements allowed beyond which two datasets are extracted
                instead of one. Avoids creating too many subsets when lat/lon
                are not continuous.

            masked (bool): mask the data out of the geometry boundary.
                May take additional processing time.

            as_masked_array (bool): return result as a numpy masked array if
                True (default), or xarray DataArray if set to False.

        Return:
            data subset fitting the clip area, as a xarray DataArray or numpy
            masked array, depending on ``as_masked_array`` value.
        """
        (lonmin, latmin, lonmax, latmax) = geometry.bounds

        # get spatial dimensions
        spatial_dims = list(self.cfarray.coords['lat'].dims)
        for d in self.cfarray.coords['lon'].dims:
            if d not in spatial_dims:
                spatial_dims.append(d)

        # coarse search on geometry bounding box - lat/lon need to be
        # broadcasted to the array spatial shape
        lat = self.cfarray.lat.cb.isel(
            broadcast_coords=[self.cfarray.coords[_] for _ in spatial_dims],
            as_masked_array=False)
        lon = self.cfarray.lon.cb.isel(
            broadcast_coords=[self.cfarray.coords[_] for _ in spatial_dims],
            as_masked_array=False)

        area_data = lat.where((
                (lon >= lonmin) & (lon <= lonmax) &
                (lat >= latmin) & (lat <= latmax)
        ))
        inside = ~area_data.to_masked_array().mask
        if continuity_threshold > 0:
            inside = binary_dilation(inside, iterations=continuity_threshold)
        slices = find_objects(label(inside)[0])

        # return all extracted subsets
        subsets = []
        for sl in slices:
            dslices = {d: sl[i] for i, d in enumerate(spatial_dims)}
            if 'index' in kwargs:
                if set(spatial_dims).intersection(kwargs['index'].keys()):
                    raise ValueError(
                        'slicing on spatial dimensions can not be used when '
                        'clipping data'
                    )
                dslices.update(kwargs.pop('index'))
            subsets.append(
                self.cfarray.cb.isel(dslices, **kwargs))

            if masked:
                # @TODO
                raise NotImplementedError

        return subsets



# -*- coding: utf-8 -*-
"""
Class for the swath feature
"""
import typing as T
import warnings

import xarray as xr

import cerbere
from cerbere.feature.cbasefeature import BaseFeature
from cerbere.feature.cgrid import Grid


__all__ = ['Swath']


class Swath(Grid):
    """
    Feature class for representing a swath, a two-dimensional irregular grid
    along the satellite ground track.
    """
    cdm_data_type = 'swath'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': ('X', 'Y'),
        'latitude': ('X', 'Y'),
        'time': [('X', 'Y'), ('Y',)],
        'vertical': [('Z',), ('X', 'Y')],
        'data': ('X', 'Y')
    }

    # optional axes for the feature
    cf_optional_axes = ['vertical']

    # axis naming
    cb_axis_naming = {'X': 'across_track', 'Y': 'along-track'}

    def check_feature(cls, dataset: xr.Dataset) -> bool:
        with warnings.catch_warnings(category=cerbere.CWarning):
            warnings.simplefilter("ignore")
            return Grid.check_feature(dataset)

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[BaseFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_cdm_data_type(cfdst):
            return

        # check horizontal axes are defined or at least some mapping
        if cfdst.cb.X is not None or cfdst.cb.Y is not None:
            return

        if cfdst.cb.cf_axis_dims['X'] == cfdst.cb.cf_axis_dims['Y']:
            return

        # check lat/lon/time are two-dimensional and share the same
        # dimensions
        if not set(cfdst['lon'].dims) == set(cfdst['lat'].dims) or not \
                (set(cfdst['lat'].dims) == set(cfdst['time'].dims) or
                 {cfdst.cb.cf_axis_dims['Y']} == set(cfdst['time'].dims)):
            return

        if cfdst.cb.vertical is not None and set(cfdst['lon'].dims) != set(
                cfdst.cb.vertical.dims):
            return

        if len(set(cfdst['lon'].dims)) != 2:
            return

        return Swath



# -*- coding: utf-8 -*-
"""
Base class for any `cerbere` dataset. Built as a wrapper around a
:class:`~xarray.Dataset` object.
"""
import datetime
import logging
import os
import warnings
from pathlib import Path
from typing import Any, Dict, Hashable, List, Mapping, Optional, Tuple, Union
import typing as TP

import numpy as np
import pandas as pd
import shapely.geometry
import xarray as xr

from scipy.ndimage import find_objects, label, binary_dilation

import cerbere
from cerbere.accessor.cerberize import (
    _cerberize, CB_CERBERIZED, CB_COORDS, CB_AXIS_DIMS, CB_AXIS_COORDS,
    CB_INSTANCE_DIMS, CB_INSTANCE_COORDS, guess_cf_aux_coords)
import cerbere.cfconvention as cf
import cerbere.accessor.matching as match
from cerbere.exceptions import CoordinateError
from cerbere.accessor.netcdf import to_netcdf
from cerbere.feature.cdiscretefeature import DiscreteFeature


@xr.register_dataset_accessor("cb")
class DatasetAccessor:
    """
    """
    def __init__(self, xarray_obj):
        self._orig_dataset = None
        self._cfdataset = xarray_obj

        try:
            if xarray_obj.encoding.get(CB_CERBERIZED, False):
                return
        except AttributeError:
            pass

        # cached properties
        self._cfdataset.encoding[CB_CERBERIZED] = False

        # cerberize by guessing
        self.cerberize()

    def isel(
            self,
            indexers: Any = None,
            padding: bool = False,
            broadcast_coords: List[str] = None,
            prefix: Optional[str] = None,
            **kwargs: Any):
        """

        Args:
            indexers:
            padding:
            broadcast_coords:
            prefix:
            **kwargs:

        Returns:

        """
        if padding is False and broadcast_coords is None:
            subs = self.cfdataset.isel(indexers, **kwargs)
        else:
            if broadcast_coords is not None:
                for c in broadcast_coords:
                    if c not in self.cfdataset.coords:
                        raise ValueError(
                            f'{c} is not a coordinate, broadcast along this '
                            f'dimension is forbidden.'
                        )

            isel_args = kwargs.copy()
            isel_args.update(
                {'padding': padding, 'broadcast_coords': broadcast_coords})

            subs = xr.Dataset(
                attrs=self._cfdataset.attrs,
                data_vars={
                    _: self._cfdataset[_].cb.isel(indexers, **isel_args)
                    for _ in self._cfdataset.data_vars})
            # add missing coords
            for coord in self._cfdataset.coords:
                if coord not in subs.coords:
                    subs.coords[coord] = self._cfdataset[coord].cb.isel(
                        indexers, **isel_args)

            subs.encoding.update(self._cfdataset.encoding.copy())

        # prefix the output variables
        if prefix:
            subs = subs.rename({_: f'{prefix}_{_}' for _ in subs.data_vars})

        return subs

    @property
    def cfdataset(self) -> xr.Dataset:
        """Returns a CF normalized version of the dataset."""
        if self._cfdataset is None:
            # default standardization - may not be enough for some products
            # that have more unusual none-conformity
            self.cerberize()
        return self._cfdataset

    def is_cerberized(self):
        return self._cfdataset.encoding.get(CB_CERBERIZED, False)

    def cerberize(
            self,
            dim_matching: Optional[Dict[str, str]] = None,
            coord_matching: Optional[Dict[str, str]] = None,
            attr_matching: Optional[Dict[str, str]] = None,
            axis_coordinates: Optional[Dict[str, str]] = None,
            **kwargs):
        """
        Args:
            coord_matching (dict): explicitly provides the matching between file
                native vars (keys) and their Cerbere/CF equivalent (values).
        """
        # back-up original version for encoding information
        # @TODO may have a memory imprint if some data copy occur - may be
        #  better to only save variable and global encoding
        self._orig_dataset = self._cfdataset.copy(deep=False)

        # harmonize dataset wrt cerbere CF & ACDD conventions
        self._cfdataset = _cerberize(
            self._cfdataset,
            dim_matching=dim_matching,
            coord_matching=coord_matching,
            attr_matching=attr_matching,
            axis_coordinates=axis_coordinates
        )


    @property
    def latitude(self) -> xr.DataArray:
        """"Latitude coordinate, as a DataArray"""
        return self._cfdataset['lat']

    @property
    def longitude(self) -> xr.DataArray:
        """"Longitude coordinate, as a DataArray"""
        return self._cfdataset['lon']

    @property
    def time(self) -> xr.DataArray:
        """"Time coordinate, as a DataArray"""
        return self._cfdataset['time']

    @property
    def vertical(self) -> TP.Optional[xr.DataArray]:
        """"Vertical coordinate (height or depth), as a DataArray"""
        try:
            return self._cfdataset[self.cf_coords['vertical']]
        except KeyError:
            return

    def _encoding_attr(self, value: str, category: str =None) -> str:
        if category is None:
            return self._cfdataset.encoding[value]
        return self._cfdataset.encoding[category][value]

    def _assign_coordinate(self, coord: str, data_var: xr.DataArray):
        pass

    # def _guess_cf_coord_lat(self) -> Hashable:
    #     """CF criteria for latitude (CF 1.11):
    #
    #     Variables representing latitude must always explicitly include the
    #     units attribute; there is no default value. The recommended value of
    #     the units attribute is the string degrees_north. Also accepted are
    #     degree_north, degree_N, degrees_N, degreeN, and degreesN.
    #
    #     Optionally, the latitude type may be indicated additionally by
    #     providing the standard_name attribute with the value latitude, and/or
    #     the axis attribute with the value Y.
    #
    #     Coordinates of latitude with respect to a rotated pole should be given
    #     units of degrees, not degrees_north or equivalents, because
    #     applications which use the units to identify axes would have no means
    #     of distinguishing such an axis from real latitude, and might draw
    #     incorrect coastlines, for instance.
    #     """
    #     return self._guess_cf_coord(self._cfdataset, 'lat')
    #
    # def _guess_cf_coord_lon(self) -> Hashable:
    #     """CF criteria for longitude (CF 1.11):
    #
    #     Variables representing longitude must always explicitly include the
    #     units attribute; there is no default value. The recommended value of
    #     the units attribute is the string degrees_east. Also accepted are
    #     degree_east, degree_E, degrees_E, degreeE, and degreesE.
    #
    #     Optionally, the longitude type may be indicated additionally by
    #     providing the standard_name attribute with the value longitude,
    #     and/or the axis attribute with the value X.
    #
    #     Coordinates of longitude with respect to a rotated pole should be given
    #     units of degrees, not degrees_east or equivalents, because applications
    #     which use the units to identify axes would have no means of
    #     distinguishing such an axis from real longitude, and might draw
    #     incorrect coastlines, for instance.
    #     """
    #     return self._guess_cf_coord(self._cfdataset, 'lon')
    #
    # def _guess_cf_coord_time(self) -> Hashable:
    #     """
    #     CF criteria for time coordinate (CF 1.11):
    #
    #     Variables representing reference time must always explicitly include
    #     the units attribute; there is no default value. The units attribute
    #     takes a string value formatted as per the recommendations in the
    #     [UDUNITS] package. The following excerpt from the UDUNITS documentation
    #      explains the time unit encoding by example:
    #
    #     "The specification seconds since 1992-10-8 15:15:42.5 -6:00 indicates
    #     seconds since October 8th, 1992 at 3 hours, 15 minutes and 42.5
    #     seconds in the afternoon in the time zone which is six hours to the
    #      west of Coordinated Universal Time (i.e. Mountain Daylight Time).
    #     The time zone specification can also be written without a colon using
    #     one or two digits (indicating hours) or three or four digits
    #     (indicating hours and minutes)."
    #
    #     The acceptable units for time are given by the UDUNITS package
    #      [UDUNITS]. The most commonly used of these strings (and their
    #      abbreviations) includes day (d), hour (hr, h), minute (min) and second
    #      (sec, s). Plural forms are also acceptable.
    #
    #     The reference date/time string (appearing after the identifier since)
    #     is required. It may include date alone, or date and time, or date, time
    #     and time zone. If the time zone is omitted the default is UTC, and if
    #     both time and time zone are omitted the default is 00:00:00 UTC.
    #
    #     A reference time coordinate is identifiable from its units string alone.
    #
    #     Optionally, the time coordinate may be indicated additionally by
    #     providing the standard_name attribute with an appropriate value, and/or
    #     the axis attribute with the value T.
    #
    #     Returns:
    #
    #     """
    #     return self._guess_cf_coord(self._cfdataset, 'time')
    #
    # def _guess_cf_coord_vertical(self) -> Hashable:
    #     return self._guess_cf_coord(self._cfdataset, 'vertical')

    @staticmethod
    def _verify_cf_axes(dst: xr.Dataset, axes: TP.Dict[str, TP.Hashable]):
        """To be a CF axis, the corresponding coordinate should be the only
        one depending on the axis dimension"""
        for axis in axes:
            # verify there is no other coordinate depending on this axis
            for coord in dst.cb.cf_coords:
                if coord.cb.axis != axis and axis in coord.dims:
                    raise cerbere.BadAxisError(
                        f'The axis dimension {axis} is used by different '
                        f'coordinate variables and therefore not be an axis '
                        f'dimension')
    #
    # def _guess_coord(self, axis: str) -> str:
    #     """Guess which variable is a given axis coordinate"""
    #     for c in self._cfdataset.coords:
    #         attrs = self._cfdataset.coords[c].attrs
    #         if 'axis' in attrs and attrs['axis'] == axis:
    #             return self._cfdataset.coords[c].name
    #     logging.debug(f'No {axis} coordinate found')
    #
    #     # try in variables
    #     for c in self._cfdataset.data_vars:
    #         attrs = self._cfdataset.data_vars[c].attrs
    #         if 'axis' in attrs and attrs['axis'] == axis:
    #             return self._cfdataset.data_vars[c].name
    #     logging.debug(f'No {axis} coordinate found')
    #
    # def _guess_t_coord(self) -> str:
    #     """Guess which variable is a T axis coordinate"""
    #     return self._guess_coord('T')
    #
    # def _guess_x_coord(self) -> str:
    #     """Guess which variable is a X axis coordinate"""
    #     x_coord = self._guess_coord('X')
    #     if x_coord is not None:
    #         return x_coord
    #
    #     if 'lon' in self._cfdataset.coords and \
    #             len(self._cfdataset.lon.dims) < 2:
    #         return 'lon'
    #
    # def _guess_y_coord(self) -> str:
    #     """Guess which variable is a Y axis coordinate"""
    #     y_coord = self._guess_coord('Y')
    #     if y_coord is not None:
    #         return y_coord
    #
    #     if 'lat' in self._cfdataset.coords and \
    #             len(self._cfdataset.lon.dims) < 2:
    #         return 'lat'
    #
    # def _guess_z_coord(self) -> str:
    #     """Guess which variable is a Z axis coordinate"""
    #     return self._guess_coord('Z')

    def set_as_z(self, fieldname: str, positive=None):
        """Set a field as representing the Z coordinate"""
        # add Z axis CF attribute
        self._cfdataset[fieldname].attrs['axis'] = 'Z'
        if positive is not None:
            if 'positive' in self._cfdataset[fieldname].attrs:
                logging.warning(
                    'positive attribute already defined and will be overriden '
                    'by the new value'
                )
                self._cfdataset[fieldname].attrs['positive'] = positive

        # remove previous Z axis (only one coordinate can be Z axis)
        for v in self._cfdataset.variables:
            if ('axis' in self._cfdataset[v].attrs and
                    self._cfdataset[v].attrs['axis'] == 'Z'):
                self._cfdataset[v].attrs.pop('axis')
                break

        # set as coordinate
        self._cfdataset = self._cfdataset.set_coords([fieldname])

    def __str__(self):
        result = 'Dataset: \nCF Spatiotemporal Dims :\n'
        for dim, dsize in self.cf_dims.items():
            result = result + '   .\t{} : {}\n'.format(dim, dsize)

        result = result + 'Other Dims :\n'
        for dim, dsize in self._cfdataset.dims.items():
            if dim in self.cf_dims:
                continue
            result = result + '   .\t{} : {}\n'.format(dim, dsize)

        result = result + 'CF Spatiotemporal Coordinates :\n'
        # TODO replace geocoordnames
        for coord in self.geocoordnames:
            result = result + '   .\t{}  ({})\n'.format(
                coord,
                ', '.join(
                    [(str(k) + ': ' + str(v))
                     for k, v in self._cfdataset[coord].sizes.items()]))

        result = result + 'Other Coordinates :\n'
        for coord in self.coordnames:
            if coord in self.geocoordnames:
                continue
            result = result + '   .\t{}  ({})\n'.format(
                coord,
                ', '.join(
                    [(str(k) + ': ' + str(v))
                     for k, v in self._cfdataset[coord].sizes.items()]
                )
            )

        result = result + 'Data :\n'
        for field in self._cfdataset.data_vars:
            result = result + '   .\t{}  ({})\n'.format(
                field,
                ', '.join(
                    [(str(k) + ': ' + str(v))
                     for k, v in self._cfdataset[field].sizes.items()]
                )
            )

        result = result + 'Global Attributes :\n'
        for attr in self.cfdataset.attrs:
            result += f'   .\t{attr}\t{str(self.cfdataset.attrs[attr])}\n'

        return result

    def is_empty(self) -> bool:
        """Returns True if the dataset object contains no data yet"""
        return self._cfdataset.equals(xr.Dataset())

    @property
    def basename(self) -> str:
        """Return the basename of the file storing the dataset"""
        return Path(self.url).name

    @property
    def filesize(self) -> int:
        """The dataset file size, in octets"""
        return os.path.getsize(self.url)

    @property
    def file_creation_date(self) -> 'datetime.datetime':
        """The date the dataset file was generated"""
        return datetime.datetime.fromtimestamp(os.path.getctime(self.url))

    def has_coord(self, coord: Hashable) -> bool:
        """Return True if the coordinate field is defined"""
        return coord in self._cfdataset.coords

    # @property
    # def dimnames(self) -> Tuple[Hashable]:
    #     """Tuple of the dataset's dimension names"""
    #     return tuple(self.cfdataset.sizes.keys())
    #
    # @property
    # def dimsizes(self) -> Tuple[int]:
    #     """Tuple of the dataset's dimension sizes"""
    #     return tuple(self.cfdataset.dims.values())

    @property
    def _varnames(self) -> List[Hashable]:
        """List of the names of all the fields (including coordinates) of the
        dataset.
        """
        return list(self._cfdataset.variables.keys())

    @property
    def cf_coords(self) -> TP.Dict[str, str]:
        """Return the spatiotemporal coordinate variables

        Return:
          dict: a dictionary where keys are the CF spatiotemporal coordinates (
          among longitude, latitude, vertical, time) and values are the
          coordinate names for these axes.
        """
        if not self.is_cerberized():
            self.cerberize()
        return self._cfdataset.encoding.get(CB_COORDS)

    @property
    def cf_axis_coords(self) -> TP.Dict[str, str]:
        """Return the spatiotemporal axis coordinate variables.

        The axis coordinates may be different from the coordinates,
        for instance in curvilinear grids having projection X and Y
        axis coordinates.

        Return:
          dict: a dictionary where keys are the CF axes (X, Y, Z, T) and
            values the coordinate names for these axes.
        """
        if not self.is_cerberized():
            self.cerberize()
        return self._cfdataset.encoding.get(CB_AXIS_COORDS)

    @staticmethod
    def _find_axis_coord(
            dst: xr.Dataset, axis: str) -> TP.Optional[TP.Hashable]:
        """Return the name of the spatiotemporal coordinate associated
        with a given CF axis, based on the CF axis attribute.

        Args:
            dst: the dataset in which to search the axis coordinate
            axis: axis name among 'X', 'Y', 'T', 'Z'
        """
        if axis not in cf.CF_AXIS:
            raise cerbere.AxisError(
                '{} is not a geolocation coordinate'.format(axis))

        axis_coord = [
            coord for coord in dst.variables
            if dst[coord].attrs.get('axis') == axis]

        if len(axis_coord) == 0:
            raise cerbere.MissingAxisError(
                f'No coordinate found for axis: {axis}')
        if len(axis_coord) != 1:
            raise cerbere.BadAxisError(
                f'Too many coordinates found for axis: {axis}: {axis_coord}. '
                f'There can be only one coordinate for a given axis. This file '
                f'or dataset should be fixed first.')

        return axis_coord[0]

    def rename_cf_axis_coord(self, axis: str, name: str):
        """
        Rename an axis coordinate.

        Update cerbere internals at the same time to maintain encoding
        consistency.

        Args:
            axis: axis coordinate to rename, among 'X', 'Y', 'Z' or 'T'
            name: new name for the axis coordinate
        """
        if axis not in cf.CF_AXIS:
            raise cerbere.AxisError(f'Illegal axis {axis}')
        if axis in self.cf_axis_coords:
            raise cerbere.AxisError(f'{axis} is not an axis coordinate')
        prev_name = self.cf_axis_coords.pop(axis)
        self.cf_axis_coords[axis] = name
        if prev_name in self.cf_coords.values():
            coord = {self.cf_coords[_]: _ for _ in self.cf_coords}[prev_name]
            self.cf_coords[coord] = name

    def rename_cf_axis_dim(self, axis: str, name: str):
        """
        Rename an axis dimension.

        Update cerbere internals at the same time to maintain encoding
        consistency.

        Args:
            axis: axis dimension to rename, among 'X', 'Y', 'Z' or 'T'
            name: new name for the axis dimension
        """
        self.cf_axis_dims[axis] = name

    def set_axis(self, name: str, axis: str):
        """Set a variable as a CF axis.

        Args:
            name: name of the variable to set as an axis
            axis: value of the axis
        """
        if axis not in cf.CF_AXIS:
            raise cerbere.AxisError(f'Illegal axis {axis}')

        # verify there is no variable with this axis already
        axis_vars = set([
            _ for _ in self._cfdataset.variables
            if self._cfdataset[_].attrs.get('axis') == axis]) - {name}

        if len(axis_vars) > 0:
            logging.warning(
                f'The following variable(s) were set as {axis} axis: '
                f'{axis_vars}. They will be unset by this action.'
            )
            if axis in self.cf_axis_coords:
                self.cf_axis_coords.pop(axis)
            for v_axis in axis_vars:
                self._cfdataset[v_axis].attrs.pop('axis')
                if v_axis in self._cfdataset.coords:
                    self._cfdataset = self._cfdataset.reset_coords(v_axis)

        self._cfdataset[name].attrs['axis'] = axis
        self.cf_axis_coords[axis] = name
        self._cfdataset = self._cfdataset.set_coords([name])

        # re-cerberize to update internal encoding
        self._cfdataset = _cerberize(self._cfdataset)

    def _cf_axis(self, axis: str):
        """Return the corresponding dimension name for a given axis"""
        if not self.is_cerberized():
            self.cerberize()
        return self._cfdataset.encoding.get('_cb_axes').get(axis)

    @property
    def X(self) -> Optional[xr.DataArray]:
        """Return the X axis coordinate variable"""
        if self.cf_axis_coords.get('X') is None:
            return
        return self._cfdataset[self.cf_axis_coords['X']]

    @property
    def Y(self) -> Optional[xr.DataArray]:
        """Return the Y axis coordinate variable"""
        if self.cf_axis_coords.get('Y') is None:
            return
        return self._cfdataset[self.cf_axis_coords['Y']]

    @property
    def T(self) -> Optional[xr.DataArray]:
        """Return the T (time) axis coordinate variable"""
        return self._cfdataset[self.cf_axis_coords['T']]

    @property
    def Z(self) -> Optional[xr.DataArray]:
        """Return the Z (vertical) axis coordinate variable"""
        return self._cfdataset[self.cf_axis_coords['Z']]

    @property
    def cf_axis_dims(self) -> Mapping[str, str]:
        """Return the mapping from CF spatiotemporal axis (X, Y, Z,
        T) to the dimension names in the dataset
        """
        return self._cfdataset.encoding[CB_AXIS_DIMS]

    @property
    def cf_instance_dims(self) -> List[str]:
        """Return the instance axis dimensions of a feature collection
        """
        return self._cfdataset.encoding[CB_INSTANCE_DIMS]

    @property
    def cf_instance_coords(self) -> List[str]:
        """Return the instance axis coordinates of a feature collection
        """
        return self._cfdataset.encoding[CB_INSTANCE_COORDS]

    @property
    def cf_dims(self) -> Mapping[str, int]:
        """Return the mapping from CF spatiotemporal dimension names to lengths.
        """
        return {
            _: self._cfdataset.sizes[_]
            for _ in list(self._cfdataset.dims)
            if _ in self._cfdataset.encoding[CB_AXIS_DIMS].values()}

    @property
    def is_collection(self):
        return len(self.cf_instance_dims) > 0

    def set_z(self, values, ztype='depth'):
        """Set the depth(s)

        Args:
            values (:class:`numpy.ma.MaskedArray` or :class:`Field`): depth
                values, as a array of floats or a Field object

            ztype (string): type is depth (positive down) or height (positive
                up)
        """
        coord = self._make_coord(
            'z', values, 'm', ztype, ztype
        )
        coord.attrs.update(
            {'depth': {'positive': 'down'},
             'height': {'positive': 'up'}
             }[ztype]
        )
        return coord

    def clip(
            self,
            geometry: shapely.geometry.Polygon,
            masked: bool = False,
            continuity_threshold: int = 0,
            **kwargs) -> List[xr.Dataset]:
        """Return the dataset subsets within a geographical area.

        Extract the subset(s) of the Dataset including the whole area geometry,
        defined as a polygon of lon/lat coordinates.

        Any keywords from :meth:`extract()` can be applied here too.

        Args:
            geometry (shapely.geometry.Polygon): boundaries of the clipped data,
                as a polygon of lon, lat coordinates.

            continuity_threshold (int): maximum number of non-contiguous
                measurements allowed beyond which two datasets are extracted
                instead of one. Avoids creating too many subsets when lat/lon
                are not continuous.

            masked (bool, optional): mask the data out of the geometry boundary.
                May take additional processing time.

        Return:
            MaskedArray: array of data read. Array type is the same as the
                storage type.
        """
        # coarse search on geometry bounding box
        (lonmin, latmin, lonmax, latmax) = geometry.bounds

        # get spatial dimensions
        spatial_dims = list(self.cfdataset.coords['lat'].dims)
        for d in self.cfdataset.coords['lon'].dims:
            if d not in spatial_dims:
                spatial_dims.append(d)

        # coarse search on geometry bounding box
        lat = self.cfdataset.cb.latitude.cb.isel(
            broadcast_coords=[self.cfdataset.coords[_] for _ in spatial_dims],
            as_masked_array=False)
        lon = self.cfdataset.cb.longitude.cb.isel(
            broadcast_coords=[self.cfdataset.coords[_] for _ in spatial_dims],
            as_masked_array=False)

        area_data = lat.where((
                (lon >= lonmin) & (lon <= lonmax) &
                (lat >= latmin) & (lat <= latmax)
        ))
        inside = ~area_data.to_masked_array().mask
        if continuity_threshold > 0:
            inside = binary_dilation(inside, iterations=continuity_threshold)
        slices = find_objects(label(inside)[0])

        # return all extracted subsets
        subsets = []
        for sl in slices:
            dslices = {d: sl[i] for i, d in enumerate(spatial_dims)}
            if 'index' in kwargs:
                if set(spatial_dims).intersection(kwargs['index'].keys()):
                    raise ValueError(
                        'slicing on spatial dimensions can not be used when '
                        'clipping data'
                    )
                dslices.update(kwargs.pop('index'))
            subsets.append(self.isel(dslices, **kwargs))

            if masked:
                raise NotImplementedError

        return subsets

    def get_location_by_value(self, field, value, **kwargs):
        """Return the indices and geolocation of a given value.

        If the value is not unique, the first one found is returned.
        """
        # TODO remove or update
        if not np.isscalar(value):
            raise ValueError('value must be a scalar')

        loc = np.where(self.get_values(field, **kwargs) == value)
        loc = {dim: loc[i][0]
               for i, dim in enumerate(self.get_field(field).dimnames)}

        geoloc = (np.asscalar(self.get_lon(index=loc)),
                  np.asscalar(self.get_lat(index=loc)))

        return loc, geoloc,

    def closest_spatial_location(self, lon, lat):
        """Get closest dataset lat/lon location to given coordinates.

        Use pythagorian differences on lat/lon values so take the result with
        caution.

        Returns:
            A tuple with the indices and lon/lat of the closest point found.
        """
        # TODO remove or update
        d_lat = self.get_lat() - lat
        d_lon = self.get_lon() - lon

        if len(d_lat.shape) == 2:
            r2 = d_lat ** 2 + d_lon ** 2
            loc = np.where(r2 == np.min(r2))
            loc = {dim: loc[i][0]
                   for i, dim in enumerate(self.get_field_dims('lat'))}
            geoloc = (np.asscalar(self.get_lon(index=loc)),
                      np.asscalar(self.get_lat(index=loc)))
            return loc, geoloc,

        else:
            # @TODO
            raise NotImplementedError

    @property
    def bbox(self) -> 'shapely.geometry.Polygon':
        """The bounding box, e.g. the south/north and east/west
        limits of the feature.
        """
        if 'bbox' not in self.cfdataset.attrs:
            # calculate bbox automatically if not explicitely defined
            self.bbox = 'auto'
        return self.cfdataset.attrs['bbox']

    @bbox.setter
    def bbox(self,
             bbox: Union[str, tuple, shapely.geometry.Polygon] = 'auto'):
        """Set the bounding box of the feature

        Args:
            :class:`shapely.geometry.Polygon`: the bounding box. If None, it is
                estimated from the feature lat/lon.
        """
        if isinstance(bbox, tuple):
            warnings.warn(
                'bbox as a tuple is deprecated. Use shapely geometry.',
                FutureWarning
            )
            self.cfdataset.attrs['bbox'] = shapely.geometry.box(bbox)
        elif isinstance(bbox, shapely.geometry.Polygon):
            self.cfdataset.attrs['bbox'] = bbox
        elif bbox == 'auto':
            # look first at global attributes
            attrs = self.cfdataset.attrs
            limits = []
            for limit in ['lonmin', 'latmin', 'lonmax', 'latmax']:
                # check various geographical attribute styles
                for att in match.BBOX_CORNERS[limit]:
                    fatt = next((x for x in attrs if x == att), None)
                    if fatt is not None:
                        limits.append(attrs[fatt])
            if len(limits) == 4:
                self.cfdataset.attrs['bbox'] = shapely.geometry.box(*limits)

            # estimate from the lat/lon
            elif self.latitude is not None and self.longitude is not None:
                lats = self.latitude
                lons = self.longitude
                self.cfdataset.attrs['bbox'] = shapely.geometry.box(
                    lons.min(),
                    lats.min(),
                    lons.max(),
                    lats.max()
                )
            else:
                logging.warning('Could not infer bbox')
                self.cfdataset.attrs['bbox'] = None
        else:
            raise TypeError('Wrong type for bbox: {}'.format(type(bbox)))

    @property
    def wkt_bbox(self) -> str:
        """The bounding box in WKT format."""
        if self.bbox is not None:
            return self.bbox.wkt

    def save(self,
             path: Union[str, Path],
             profile: str = 'default_saving_profile.yaml',
             writer: str = None,
             force_profile: bool = False,
             keep_src_encoding: bool = False,
             **kwargs
             ):
        """
        Args:
            path (str, optional): save to a new file, whose path is provided in
                this argument. Raise an exception if the file already exists.
            profile (str, optional): path to a specific format profile to
                apply before saving (or default formatting profile is used).
            force_profile (bool, optional): force profile attribute values to
                supersede existing ones in dataset attributes.
            keep_src_encoding (bool): keep original dtype, _FillValue
                and scaling if any (through `add_offset` or `scale_factor`
                attributes) as in the source data.
        """
        if writer is not None:
            ds = cerbere.reader_class(writer).preprocess(self.cfdataset)
        else:
            ds = self.cfdataset.copy(deep=False)

        to_netcdf(
            path, ds, profile=profile,
            force_profile=force_profile,
            keep_src_encoding=keep_src_encoding, **kwargs)

    @property
    def collection_id(self) -> str:
        """return the identifier of the product collection"""
        if 'id' in self._cfdataset.attrs:
            return self._cfdataset.attrs['id']

    @property
    def naming_authority(self) -> str:
        if 'Conventions' in self._cfdataset.attrs:
            return self._cfdataset.attrs['Conventions']

    @property
    def product_version(self):
        """Return the product version"""
        if 'product_version' in self._cfdataset.attrs:
            return self._cfdataset.attrs['product_version']

    @property
    def url(self) -> Path:
        return Path(self._cfdataset.encoding.get('source'))

    @url.setter
    def url(self, value: Union[str, Path]):
        self._cfdataset.encoding['source'] = value

    def _time_coverage_from_coord(self):
        """Infer the dataset time coverage from its time coordinate.

        May cost some computation time and memory as the whole time
        coordinate will be loaded. It may be smarter to address this in a
        reader for a given dataset.
        """
        logging.warning(
            'loading the time coordinate in memory to extract the start and '
            'end time coverage')

        times = self._cfdataset.time
        if times.count() == 0:
            logging.warning('No valid time in dataset.')
            return

        # cache result
        start = pd.to_datetime(times.min(skipna=True).data)
        end = pd.to_datetime(times.max(skipna=True).data)
        # if len(times.dims) == 0:
        #     start = pd.to_datetime(times.min().data)
        #     end = pd.to_datetime(times.max().data)
        # else:
        #     start = pd.to_datetime(times[times.notnull()].min().values)
        #     end = pd.to_datetime(times[times.notnull()].max().values)

        self._cfdataset.attrs['time_coverage_start'] = start.to_pydatetime()
        self._cfdataset.attrs['time_coverage_end'] = end.to_pydatetime()

    def _time_coverage(
            self,
            boundary='time_coverage_start') -> datetime.datetime:
        if self._cfdataset.attrs.get(boundary) is None:
            # compute from time coordinate - may take some time
            self._time_coverage_from_coord()
        return self._cfdataset.attrs[boundary]

    @property
    def time_coverage_start(self) -> datetime.datetime:
        """Returns the first measurement time in the data.

        If the information is not available as an internal attribute, it will be
        computed (and cached) which may take some computation time as it
        required to read the values of the time coordinate.
        """
        return self._time_coverage('time_coverage_start')

    @time_coverage_start.setter
    def time_coverage_start(self, value: datetime.datetime):
        self._cfdataset.attrs['time_coverage_start'] = value

    @property
    def time_coverage_end(self) -> datetime.datetime:
        """Returns the last measurement time in the data.

        If the information is not available as an internal attribute, it will be
        computed (and cached) which may take some computation time as it
        required to read the values of the time coordinate.
        """
        return self._time_coverage('time_coverage_end')

    @time_coverage_end.setter
    def time_coverage_end(self, value: datetime.datetime):
        self._cfdataset.attrs['time_coverage_end'] = value

    @property
    def geospatial_bounds(self) -> str:
        return self._cfdataset.attrs.get('geospatial_bounds', None)

    @property
    def featureType(self) -> str:
        """the CF feature type of the data contained in the dataset"""
        return self._cfdataset.attrs.get('featureType', None)

    @property
    def cdm_data_type(self) -> str:
        """the Unidata Common Data Model feature type of the data contained in
        the dataset"""
        return self._cfdataset.attrs.get('cdm_data_type', None)

    @property
    def cf_role(self) -> Dict[str, str]:
        """Returns the name and role of the coordinates holding the cf role
        keyword"""
        return {v: self._cfdataset[v].attrs.get('cf_role')
                for v in self._cfdataset.variables}

    @property
    def cf_auxiliary_coords(self) -> TP.List[str]:
        """The spatiotemporal coordinates, among longitude, latitude,
        vertical or time, which depends on more than one spatiotemporal
        dimension"""
        return guess_cf_aux_coords(self._cfdataset).keys()

    @property
    def grid_mapping(self) -> TP.Union[xr.DataArray, TP.List[xr.DataArray]]:
        """Return the grid mapping variable(s)"""
        mvars = [_ for _ in self._cfdataset.data_vars
                 if self._cfdataset[_].attrs.get('grid_mapping_name')]

        if len(mvars) == 0:
            return

        if len(mvars) == 1:
            return self._cfdataset[mvars[0]]

        return [self._cfdataset[_] for _ in mvars]

    @property
    def grid_mapping_name(self) -> TP.Union[str, TP.List[str]]:
        """Return the grid mapping name"""
        mvars = self.grid_mapping
        if mvars is None:
            return

        if isinstance(mvars, list):
            return [
                _.attrs.get('grid_mapping_name') for _ in mvars]

        return mvars.attrs.get('grid_mapping_name')



# -*- coding: utf-8 -*-
"""
:mod:`~cerbere.dataset` class for GHRSST netcdf files
"""
from __future__ import absolute_import, division, print_function

import datetime
import uuid

import cftime
import netCDF4
import numpy
import xarray as xr

from cerbere.dataset import ENCODING, Encoding
from cerbere.dataset.dataset import CDM_TYPES, OpenMode
from cerbere.dataset.ncdataset import NCDataset

from ..cfconvention import DEFAULT_TIME_UNITS

# common matching dimensions and fields in netCDF files
GHRSST_DIM_MATCHING = {
    'time': u'time',
    'lon': u'lon',
    'lat': u'lat',
    'x': u'x',
    'y': u'y',
    'ni': u'cell',
    'nj': u'row'
}


GHRSST_FIELD_MATCHING = {
    'time': u'time',
    'lon': u'lon',
    'lat': u'lat'
}


GHRSST_LEVELS = ['L4', 'L3P', 'L3C', 'L3U', 'L2P']


class GHRSSTNCDataset(NCDataset):
    """
    GHRSST GDS 2.x compliant NetCDF Dataset.
    """
    def __init__(self, *args, **kwargs):
        kwargs['dim_matching'] = GHRSST_DIM_MATCHING
        kwargs['field_matching'] = GHRSST_FIELD_MATCHING
        super().__init__(*args, **kwargs)

    def _open(self, **kwargs):
        # url needs to be opened first in order to guess default datamodel
        super()._open(**kwargs)

        if self._mode != OpenMode.WRITE_NEW:

            # recompose and unpack time variable if required
            self._unsplit_time_coord(
                split_time=('time', 'sst_dtime'), unpack=True
            )

            # normalize collection id (GDS 1.x => 2.x)
            if 'DSD_entry_id' in self.dataset.attrs:
                self.dataset.attrs['id'] = self.dataset.attrs.pop(
                    'DSD_entry_id'
                )

            # fix row/cell ordering (NAVO & AMSRE products)
            if (('NAVO' in self.get_collection_id() and
                 'VIIRS' not in self.get_collection_id()) or
                    self.get_collection_id() == 'USA-RSS-AMSRE-MW-L2-SST'):
                self.dataset.swap_dims({'row': 'cell', 'cell': 'row'})

            # fix lat/lon in MODIS  products
            # @TODO: to be checked
            if (self.get_collection_id() in [
                'JPL-L2P-MODIS_A',
                'MODIS_A-JPL-L2P-v2014.0',
                'JPL-L2P-MODIS_T',
                'MODIS_T-JPL-L2P-v2014.0'
            ]):
                # some lats have nan
                self.dataset['lat'].values = numpy.ma.fix_invalid(
                    self.dataset['lat'].values,
                    copy=False
                )
                # and corresponding lons have 0.0.... !
                self.dataset['lon'].values = numpy.ma.array(
                    self.dataset['lon'].values,
                    mask=self.dataset['lat'].values.mask,
                    copy=False
                )

            # fix lat/lon in NAVO  products
            if 'NAVO' in self.get_collection_id():
                # NAVO lat/lon contains 0 values
                for coord in ['lat', 'lon']:
                    self.dataset[coord].values = numpy.ma.masked_values(
                        self.dataset[coord].values,
                        0,
                        copy=False
                    )

            # @TODO
            #if datamodel is None:
            #    self._feature_type = self.__guess_feature()

    def _unsplit_time_coord(self, split_time, unpack=True):
        """Build a single time fields from two separate fields.

        This address the case of some dataset where each time value is the sum
        of a reference time and an offset. For instance, in GHRSST products,
        the actual pixel time is the sum of``time`` and ``sst_dtime`` fields.
        """
        time_var, dtime_var = split_time

        dataset = self.dataset
        if dtime_var not in dataset.data_vars:
            raise ValueError(
                'Unknown {} field in data fields'.format(dtime_var)
            )
        if time_var not in dataset.coords:
            raise ValueError(
                'Unknown {} field in coordinate fields'.format(time_var)
            )

        dtime = dataset[dtime_var]
        fillvalue = dataset[dtime_var].encoding['_FillValue']
        if 'units' in dtime.attrs:
            dunit = dtime.attrs['units']
        elif 'units' in dtime.encoding:
            dunit = dtime.encoding['units']
        else:
            raise ValueError('missing unit for sst_dtime')
        dtime.attrs = dataset[time_var].attrs
        dtime.encoding = dataset[time_var].encoding
        dtime.encoding[ENCODING][Encoding.IO_FILLVALUE] = fillvalue
        reference = numpy.datetime64(
            dataset[time_var].values[0]
        )
        if unpack:
            if dunit not in ['second', 'seconds']:
                raise TypeError(
                    'Can not process sst_dtime in {}'.format(dunit)
                )
            dtime.values = numpy.ma.array(
                dtime.values, dtype='timedelta64[s]', copy=False
            ) + reference
        else:
            dtime.encoding['add_offset'] = reference

        # remove dummy time dimension
        dataset = dataset.squeeze(dim='time')

        # replace with single time field
        dataset = dataset.drop_vars([time_var])
        dataset = dataset.rename({dtime_var: 'time'})
        dataset = dataset.set_coords([time_var])
        if 'comment' in dataset['time'].attrs:
            dataset['time'].attrs.pop('comment')
        dataset['time'].attrs['long_name'] = 'measurement time'

        # encoding
        dataset['time'].encoding[ENCODING] = {}
        dataset['time'].encoding[ENCODING][Encoding.M_MASK] = \
            numpy.isnat(dataset['time'].values)
        dataset['time'].encoding[ENCODING][Encoding.M_DTYPE] = \
            dataset['time'].dtype

        self.dataset = dataset

    def _convert_format(self,
            profile='ghrsst_saving_profile.yaml',
            force_profile=False,
            **kwargs):
        """Implement specific formatting rules to a dataset.

        Used before saving the dataset to match some specific format
        requirements when writing the dataset on file.
        """
        dataset = super()._convert_format(
            profile=profile, force_profile=force_profile, **kwargs)

        # split time
        if self.__get_level() in set(GHRSST_LEVELS) - set(['L4']):
            reftime = numpy.datetime64(
                self.time_coverage_start.strftime('%Y-%m-%d')
            )
            dtimes = self.get_times() - reftime
            timeattrs=dataset['time'].attrs.copy()
            dataset = dataset.drop('time')
            if self._format == 'NETCDF3_CLASSIC':
                dataset['sst_dtime'] = xr.DataArray(
                    data=dtimes.astype(numpy.int32),
                    dims=dataset['sea_surface_temperature'].dims,
                    coords=dataset.coords
                )
                dataset['sst_dtime'].attrs['units'] = 'seconds'
                origin = numpy.datetime64(
                    cftime.num2date(0, units=DEFAULT_TIME_UNITS)
                )
                dataset['time'] = xr.DataArray(
                    data=numpy.array(
                        [int((reftime - origin).item().total_seconds())]
                    ).astype(numpy.int32),
                    dims=('time',),
                    attrs=timeattrs
                )
                dataset['time'].attrs['units'] = DEFAULT_TIME_UNITS
            else:
                dataset['sst_dtime'] = xr.DataArray(
                    data=dtimes,
                    dims=dataset['sea_surface_temperature'].dims,
                    coords=dataset.coords
                )
                dataset['sst_dtime'].encoding['units'] = 'seconds'
                dataset['time'] = xr.DataArray(
                    data=[reftime],
                    dims=('time',),
                    attrs=timeattrs
                )
                dataset['time'].encoding['units'] = DEFAULT_TIME_UNITS
            dataset['time'].attrs['long_name'] = 'reference time of sst file'
            dataset['sst_dtime'].attrs['long_name'] = \
                'time difference from reference time'

            # fill in specific GHRSST attributes
        dataset.attrs['netcdf_version_id'] = netCDF4.getlibversion()

        dataset.attrs['uuid'] = str(uuid.uuid1())
        dataset.attrs['date_modified'] = datetime.datetime.now()
        dataset.attrs['date_issued'] = dataset.attrs['date_modified']
        dataset.attrs['date_metadata_modified'] = dataset.attrs['date_modified']
        dataset.attrs['geospatial_bounds'] = self.wkt_bbox

        for k, v in dataset.data_vars.items():
            if 'time' not in v.dims:
                dataset[k] = v.expand_dims('time', axis=0).assign_coords(
                    {'time': dataset.time}
                )

        return dataset

    def __get_level(self):
        id = self.get_collection_id()
        if id is None:
            raise ValueError(
                'The dataset object is missing the required id attribute'
            )
        for _ in GHRSST_LEVELS:
            if _ in id:
                return _
        raise ValueError(
            'Could not determine the product type for {}'
            .format(self.get_collection_id())
        )

    def __guess_feature(self):
        """Return the detected feature type"""
        attrs = self.read_global_attributes()
        if 'cdm_data_type' in attrs:
            val = self.read_global_attribute('cdm_data_type')
            if val not in CDM_TYPES:
                raise ValueError('Unknown cdm feature type {}'.format(val))
            return CDM_TYPES[val]

        elif 'L2' in self.get_collection_id():
            return 'Swath'
        elif ('L3' in self.get_collection_id() or
                'L4' in self.get_collection_id()):
            return 'Grid'
        else:
            raise ValueError('Could not guess feature type')

    #
    # def read_values(self, fieldname,**kwargs):
    #     return super().read_values(fieldname, **kwargs)

#
#         # ==========================================================
#         try:
#             if fieldname == 'time' and \
#                     'sst_dtime' in self.source().variables:
#                 ncvarname = 'sst_dtime'
#             else:
#                 ncvarname = fieldname
# #             if fieldname not in ['lat', 'lon']:
# #                 # GHRSST products have a time dimension which is not part of
# #                 # cerbere data model
# #                 if slices is not None and len(slices) == 2 and\
# #                         'bnds' not in fieldname:
# #                     slices.insert(0, 0)
#             if self.__is_reversed():
#                 # case where row and cells are reversed
#                 if slices:
#                     if len(slices) == 2:
#                         newslices = [slices[1], slices[0]]
#                     else:
#                         newslices = slices
#                     newslices = tuple(newslices)
#                 else:
#                     newslices = slices
#                 values = super(GHRSSTNCDataset, self).read_values(ncvarname,
#                                                                   newslices)
#                 if len(values.shape) == 2:
#                     values = numpy.ma.transpose(values)
#             elif (fieldname == 'lon' or fieldname == 'lat') and \
#                     len(self.source().variables[fieldname].shape) == 3:
#                 # case where the lat/lon have a time dimension (ARC ATSR files)
#                 if slices is not None:
#                     new_slices = slices
#                     new_slices.insert(0, 0)
#                     values = super(GHRSSTNCDataset, self).read_values(ncvarname,
#                                                                       tuple(new_slices))
#                 else:
#                     values = super(GHRSSTNCDataset, self).read_values(ncvarname,
#                                                                       slices)
#                     if values.ndim == 3:
#                         values = values[0, :, :]
#             else:
#                 values = super(GHRSSTNCDataset, self).read_values(
#                     ncvarname,
#                     slices)
#             if self.__is_reversed()\
#                     and (fieldname == 'lon' or fieldname == 'lat'):
#                 # NAVO lat/lon contains 0 values
#                 if 'NAVO' in self.get_collection_id():
#                     values = numpy.ma.masked_equal(values, 0, atol=0.0001,
#                                                    copy=False)
#             # homogenize longitudes
#             if fieldname == 'lon':
#                 ind = numpy.ma.where(values >= 180.)
#                 values[ind] = values[ind] - 360.
#                 ind = numpy.ma.where(values < -180.)
#                 values[ind] = values[ind] + 360.
#             if ncvarname == 'sst_dtime':
#                 # pixel time is time + sst_dtime
#                 var = self.source().variables['time']
#                 values = var[0] + values
#             # fix invalid values
#             if (self.get_collection_id() in ['JPL-L2P-MODIS_A',
#                                              'MODIS_A-JPL-L2P-v2014.0',
#                                              'JPL-L2P-MODIS_T',
#                                              'MODIS_T-JPL-L2P-v2014.0']):
#                 # some lats have nan
#                 if fieldname == 'lat':
#                     values = numpy.ma.fix_invalid(values, copy=False)
#                 # and corresponding lons have 0.0.... !
#                 elif fieldname == 'lon':
#                     lats = self.read_values('lat', slices=slices)
#                     lats = numpy.ma.fix_invalid(lats, copy=False)
#                     values = numpy.ma.array(values, mask=lats.mask,
#                                             copy=False)
#             return values
#         except:
#             logging.error("Could not read values for fieldname: {}".format(
#                 fieldname
#             ))
#             raise
#
#



# -*- coding: utf-8 -*-
"""
Base class for a collection of features.
"""
import typing as T

import cerbere
from cerbere.feature.cbasefeature import BaseFeature
from cerbere.feature.cdiscretefeature import DiscreteFeature


class BaseCollection(DiscreteFeature):
    """
    Feature class for collection of features.
    """
    def __init__(
            self,
            dataset,
            instance_class: str = None,
            **kwargs):
        """
        Args:
            feature_class (BaseFeature) : the feature class which the
                collection is made of.

        See:
            BaseFeature
        """
        if instance_class is None:
            # try to infer from dataset
            instance_class = BaseCollection.infer_instance_class(dataset)
            if instance_class is None:
                raise ValueError(
                    'A value must be provided for ``feature_class`` arg'
                )
        self.instance_class = cerbere.feature_class(instance_class)

        self.cf_canonical_dims = self.instance_class.cf_canonical_dims
        if self.cf_instance_axis is None:
            self.cf_instance_axis = self.instance_class.cf_instance_axis

        super().__init__(dataset, instance_class=instance_class, **kwargs)

        if self.dataset.attrs.get('featureType') is not None:
            return

        if self.cf_feature_type is None:
            self.dataset.attrs['featureType'] =  (
                self.instance_class.cf_feature_type)
        else:
            self.dataset.attrs['featureType'] = self.cf_feature_type


    @staticmethod
    def infer_instance_class(dataset):
        if dataset.cb.featureType is not None:
            for feature in cerbere.helpers._feature:
                feat_class = cerbere.feature_class(feature)
                if not issubclass(feat_class, DiscreteFeature):
                    continue
                if dataset.cb.featureType == feat_class.cf_feature_type:
                    return feature

        # try to fit one of the features
        if len(dataset.cb.cf_instance_dims) > 0:
            item = dataset.cb.isel({dataset.cb.cf_instance_dims[0]: 0})
            for feature in ['Profile', 'Point']:
                feat_class = cerbere.feature_class(feature)
                if issubclass(feat_class, DiscreteFeature):
                    try:
                        #cerbere.guess_feature(item)
                        feat_class._check_feature(item)
                        return feature
                    except (cerbere.AxisError, TypeError):
                        pass

    @classmethod
    def _canonical_dims(cls, coord: str, instance_class=None) -> T.Set[str]:
        if isinstance(instance_class.cf_canonical_dims[coord], list):
            return instance_class.cf_canonical_dims[coord][0]
        return instance_class.cf_canonical_dims[coord]

    @classmethod
    def required_dims(cls, instance_class):
        return (instance_class.cf_instance_axis,
                ) + instance_class.cf_canonical_dims

    def __getitem__(self, item):
        """For iterating over the features of a collection"""
        instance_axis = self.ds.cb.cf_instance_dims[0]

        if isinstance(item, int):
            if item >= self.dataset.sizes[instance_axis]:
                raise IndexError("index out of range")

            return self.instance_class(self.dataset.isel({instance_axis: item}))
        else:
            return self.__class__(
                self.dataset.isel({instance_axis: item}))

    def __len__(self):
        instance_axis = self.instance_class.cf_instance_axis

        return self.dataset.dims[instance_axis]



# -*- coding: utf-8 -*-
"""
Feature class for the trajectory of profile observation pattern.
"""
import typing as T
import warnings

import xarray as xr

import cerbere
from cerbere.feature.cbasecollection import BaseCollection
from cerbere.feature.comdcollection import OMDCollection
from cerbere.feature.cimdcollection import IMDCollection


__all__ = ['TrajectoryProfile']


class TrajectoryProfile(BaseCollection):
    """
    Feature class for the CF / Unidata CDM observation pattern corresponding to
    Profiles along a single trajectory.

    As defined in section H.6 of CF 1.11:
    "When profiles are taken along a trajectory, one gets a collection of
    profiles called a trajectoryProfile. A data variable may contain a
    collection of such trajectoryProfile features, one feature per
    trajectory. The instance dimension in the case of a trajectoryProfile is
    also referred to as the trajectory dimension. The instance variables,
    which have just this dimension, are also referred to as trajectory
    variables and are considered to contain information describing the
    trajectories. The trajectory variables may contain missing values. This
    allows one to reserve space for additional trajectories that may be added at
    a later time, as discussed in section 9.6. TrajectoryProfiles are more
    complicated than trajectories because there are two element dimensions.
    Each trajectory has a number of profiles as its elements, and each
    profile  has a number of data from various levels as its elements. It is
    strongly recommended that there always be a variable (of any data type) with
     the profile dimension and the cf_role attribute " profile_id ", whose
     values uniquely identify the profiles."

    More specifically, this class implements the case H.6.2. Profiles along a
    single trajectory:
    "If there is only one trajectory in the data variable, there is no need
    for the trajectory dimension"

    Case H.6.1. Multidimensional array representation of trajectory profiles
    is not yet implemented in this class.

    The variable(p,o) data for element o of profile p are associated with the
    coordinate values time(p), alt(p,o), lat(p), and lon(p). If all the
    profiles have the same set of vertical coordinates, the vertical
    auxiliary coordinate variable could be one-dimensional alt(z),
    or replaced by a one-dimensional coordinate variable  z(z), provided the
    values are ordered monotonically. In the latter case, listing the
    vertical coordinate variable in the coordinates attribute is optional.

    A :class:`~cerbere.feature.trajectory.Trajectory` feature is defined by
    geolocation coordinates field: ``lon(time)``, ``lat(time)``,
    ``time(time)``, optionally ``z(time)``, all having a single geolocation
    dimension: ``time``.
    """
    cf_instance_axis = 'trajectory'
    cf_element_instance_axis = 'profile'
    cf_feature_type = 'trajectoryProfile'
    cdm_data_type = 'trajectory_profile'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': (cf_instance_axis, 'profile'),
        'latitude': (cf_instance_axis, 'profile'),
        'time': (cf_instance_axis, 'profile'),
        'vertical': [(cf_instance_axis, 'profile', 'Z'),
                     (cf_instance_axis, 'Z')],
        'data': (cf_instance_axis, 'profile', 'Z')
    }
    # optional axes for the feature
    cf_optional_axes = []

    def __init__(self, *args, **kwargs):
        kwargs['instance_class'] = 'Profile'
        super().__init__(*args, **kwargs)

    @classmethod
    def cerberize(cls, dataset: xr.Dataset, force_reorder: bool = True,
                  **kwargs):
        cfdataset = super(TrajectoryProfile, cls).cerberize(
            dataset, force_reorder, **kwargs)

        # CF specific attrs and vars
        # It is strongly recommended that there always be a trajectory
        # variable (of any data type) with the attribute
        # cf_role=”trajectory_id” attribute, whose values uniquely identify
        # the trajectories.
        # if cfdataset.cb.cf_role is None:
        #     cfdataset['trajectory'] = xr.DataArray(
        #         dims=(),
        #         attrs={'cf_role': 'trajectory_id'},
        #         data=0)


        # for _, v in cfdataset.data_vars.items():
        #     v.encoding['coordinates'] = 'time lat lon z'

        return cfdataset

    @classmethod
    def _canonical_dims(cls, coord: str, instance_class=None) -> T.Set[str]:
        return instance_class.cf_canonical_dims[coord]

    @classmethod
    def is_alias(cls, dataset):
        # @TODO clarify if and which collection
        aliases = [
            _ for _ in [IMDCollection, OMDCollection]
            if _.guess_feature(dataset) is not None]
        if len(aliases) > 0:
            return aliases[0]

    @classmethod
    def guess_feature(
            cls, dataset: xr.Dataset
    ) -> T.Optional[T.Type['TrajectoryProfile']]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        try:
            cls._check_feature(dataset, instance_class='Profile')

            # to distinguish a TrajectoryProfile from a Collection of
            # profiles, there must be a variable with `cf_role` attribute
            # having value `trajectory_id`
            if 'trajectory_id' in dataset.cb.cf_role.values():
                return cls
        except cerbere.AxisError:
            return




# -*- coding: utf-8 -*-
"""
Base class for any `cerbere` dataset. Built as a wrapper around a
:class:`~xarray.Dataset` object.
"""
import copy
import datetime
import glob
import logging
import os
import warnings
from abc import ABC
from collections import OrderedDict
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Hashable, List, Mapping, MutableMapping, Optional, Tuple, Union
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import shapely.geometry
import xarray as xr
from dateutil import parser
from scipy.ndimage import find_objects, label, binary_dilation

import cerbere.cfconvention
import cerbere.dataset as internals
import cerbere.accessor.cdataset

from ..cfconvention import CF_AUTHORITY, DEFAULT_TIME_UNITS, default_profile
from .field import Field


# opening modes
class OpenMode(Enum):
    """
    ENUM to hold the allowed values for opening resources
    """
    READ_ONLY: str = 'r'
    WRITE_NEW: str = 'w'
    READ_WRITE: str = 'r+'

# attribute for override dtype
C_DTYPE = 'cerbere_dtype'

# standard geolocation coordinates
GEOCOORDINATES = [u'time', u'lat', u'lon', u'z', u'depth', u'height']
REQUIRED_GEOCOORDINATES = [u'time', u'lat', u'lon']

# standard attributes for geolocation coordinates
AXIS = {'lat': 'Y', 'lon': 'X', 'time': 'T',
        'z': 'Z', 'depth': 'Z', 'height': 'Z'}
STANDARD_NAME = {
    'lat': 'latitude',
    'lon': 'longitude',
    'time': 'time'
}
STANDARD_Z_FIELD = ['depth', 'height', 'alt', 'z']

# CDM feature types
CDM_TYPES = {
    'swath': 'Swath',
    'grid': 'Grid'
}

# common matching dimensions and fields in netCDF files
DIM_MATCHING = {
    'time': 'time',
    'lon': 'lon',
    'longitude': 'lon',
    'lat': 'lat',
    'latitude': 'lat',
    'x': 'x',
    'y': 'y',
    'mes': 'station',
    'station': 'station',
    'ni': 'cell',
    'cell': 'cell',
    'ra_size': 'cell',
    'col': 'cell',
    'nj': 'row',
    'row': 'row',
    'az_size': 'row',
    'depth': 'z',
    'height': 'z',
    'rows': 'row',
    'columns': 'cell',
    'NUMROWS': 'row',
    'NUMCELLS': 'cell',
    'across_track': 'cell',
    'along_track': 'row',
    'TIME': 'time',
    'LATITUDE': 'lat',
    'LONGITUDE': 'lon',
    'DEPTH': 'z'
}

FIELD_MATCHING = {
    'time': 'time',
    'longitude': 'lon',
    'latitude': 'lat',
    'lon': 'lon',
    'lat': 'lat',
    'depth': 'depth',
    'height': 'alt',
    'TIME': 'time',
    'LATITUDE': 'lat',
    'LONGITUDE': 'lon',
    'DEPTH': 'depth',
    'DEPH': 'depth',
}

TIME_COVERAGE_ATTRS = {
    'date': {
        'start': 'start_date',
        'end': 'stop_date'
    },
    'time': {
        'start': 'start_time',
        'end': 'stop_time'
    },
    'time_coverage': {
        'start': 'time_coverage_start',
        'end': 'time_coverage_end'
    },
    'time_coverage2': {
        'start': 'time_coverage_start',
        'end': 'time_coverage_stop'
    },
    'meas_time': {
        'start': 'first_meas_time',
        'end': 'lastst_meas_time'
    },
}

BBOX_CORNERS = {
    'latmin': [
        'southernmost_latitude', 'geospatial_lat_min', 'south_latitude'
    ],
    'latmax': [
        'northernmost_latitude', 'geospatial_lat_max', 'north_latitude'
    ],
    'lonmin': [
        'westernmost_longitude', 'geospatial_lon_min', 'west_longitude'
    ],
    'lonmax': [
        'easternmost_longitude', 'geospatial_lon_max', 'east_longitude'
    ]
}

class Dataset(ABC):
    """
    The `cerbere` dataset base class.

    A :class:`Dataset` object is internally built by composition over
    :class:`xarray.Dataset` objects.

    A :class:`Dataset` object can be created in different ways:
      * from a file (or list of files), giving its full path or URL
      * from a xarray :class:`xarray.Dataset` object
      * from a dict, using xarray syntax
      * from another :class:`Dataset` object

    Creating a Dataset from a file:

    >>> from cerbere.dataset.dataset import Dataset
    >>> dst = Dataset('./mydatafile.nc')

    Creating a Dataset from an xarray :class:`xarray.Dataset` object:
    The :mod:`xarray` object must have latitude, longitude and time coordinates
    with valid `cerbere` names (``lat``, ``lon``, ``time``):

    >>> import xarray as xr
    >>> import numpy as np
    >>> xrobj = xr.Dataset(
        coords={
            'lat': np.arange(0,10, 0.1),
            'lon': np.arange(5,15, 0.1),
            'time': np.full((100,), np.datetime64(('2010-02-03'), dtype='D'))
            },
        data_vars={'myvar': (('time',), np.ones(100))}
        )
    >>> dst = Dataset(xrobj)

    Creating a dataset from a dictionary:
    Using the same syntax as xarray (see:
    http://xarray.pydata.org/en/stable/generated/xarray.Dataset.from_dict.html#xarray.Dataset.from_dict).
    The provided dict must have latitude, longitude and time coordinates
    with valid `cerbere` names (``lat``, ``lon``, ``time``):

    >>> dst = Dataset(
            {'time': {'dims': ('time'), 'data': [datetime(2018, 1, 1)]},
             'lat': {'dims': ('lat'), 'data': np.arange(-80, 80, 1)},
             'lon': {'dims': ('lon',), 'data': np.arange(-180, 180, 1)},
             'myvar': {'dims': ('lat', 'lon',),
                       'data': np.ones(shape=(160, 360))}
             }
        )

    >>> dst = Dataset(
            {'coords': {
                'time': {'dims': ('time'), 'data': [datetime(2018, 1, 1)],
                         'attrs': {'units': 'seconds since 2001-01-01 00:00:00'}},
                'lat': {'dims': ('lat'), 'data': np.arange(-80, 80, 1)},
                'lon': {'dims': ('lon',), 'data': np.arange(-180, 180, 1)},
            },
            'attrs': {'gattr1': 'gattr_val'},
            'dims': ('time', 'lon', 'lat'),
            'data_vars': {'myvar': {'dims': ('lat', 'lon',),
                                    'data': np.ones(shape=(160, 360))}}}
        )

    :class:`Field` objects can also be mixed in:

    >>> field = Field(
            np.ones(shape=(160, 360)),
            'myvar',
            dims=('lat', 'lon',),
            attrs={'myattr': 'attr_val'}
        )
    >>> dst = Dataset(
            {'time': {'dims': ('time'), 'data': [datetime(2018, 1, 1)]},
             'lat': {'dims': ('lat'), 'data': np.arange(-80, 80, 1)},
             'lon': {'dims': ('lon',), 'data': np.arange(-180, 180, 1)},
             'myvar': field
             }
        )

    Args:

        dataset (str, list[str], xarray.Dataset): full path to a file from
            which to read the dataset content. It can be also a list of paths or
            a regular expression, if the mapper allows to open several files at
            once. Multiple files can only be opened in READ_ONLY mode.

        mode (enum, optional): access mode ('r', 'w', 'r+') when accessing a
            file.

        view (dict, optional): a dictionary where keys are dimension names
            and values are slices. A view can be set on a file, meaning
            that only the subset defined by this view will be accessible.
            This view is expressed as any subset (see :func:`get_values`).
            For example:

            >>> view = {'time':slice(0,0), 'lat':slice(200,300),
            >>>         'lon':slice(200,300)}

        dim_matching (dict): explicitly provides the matching between file
            native dimensions (keys) and their CF/cerbere standard name
            (values).

        field_matching (dict): explicitly provides the matching between file
            native fields (keys) and their CF/cerbere standard name (values).

        kwargs (dict): any argument to be passed on to the xarray
            open_dataset or open_mfdataset function.

    """
    def __init__(self,
                 dataset: Union[str, Path, xr.Dataset, 'Dataset', dict] = None,
                 mode: OpenMode = OpenMode.READ_ONLY,
                 dim_matching: Dict[str, str] = DIM_MATCHING,
                 field_matching: Dict[str, str] = FIELD_MATCHING,
                 attr_matching: Optional[Dict[str, str]] = None,
                 as_z: Optional[str] = None,
                 format: str = None,
                 attrs: Optional[Dict[str, str]] = None,
                 **kwargs):
        """
        """
        if 'view' in kwargs:
            raise DeprecationWarning('"view" keyword is deprecated and will '
                                     'not be taken into account')

        if dataset is None and 'data_vars' not in kwargs:
            raise ValueError('an url or dataset object must be provided')

        object.__init__(self)

        self.dataset = None

        # memorize opening arguments
        self._mode = OpenMode(mode)
        self._format = format
        self._url = None

        if isinstance(dataset, (str, list, Path)):
            if mode in ['w', 'rw']:
                self.dataset = xr.open_dataset(
                    dataset, mode=mode, format=format, **kwargs)
                return

            self.dataset = cerbere.open_dataset(dataset, **kwargs)

        elif isinstance(dataset, xr.Dataset):
            self.dataset = dataset

        elif isinstance(dataset, dict) or dataset is None:
            # dataset provided as a xarray compliant dictionary or xarray
            # dataset creation keywords
            if dataset is not None and 'data_vars' in kwargs:
                raise ValueError(
                    'data fields can not be provided both as main argument'
                    ' and with data_vars keyword'
                )
            if dataset is None and 'data_vars' not in kwargs:
                raise ValueError(
                    'either dataset or data_vars argument must be provided'
                )

            def data_as_tuple(arg):
                if not isinstance(arg, dict) or len(arg) == 0:
                    raise TypeError('Badly formatted input: {}'.format(arg))
                return isinstance(arg[next(iter(arg))], tuple)

            if dataset is None or data_as_tuple(dataset):
                # xarray from classic arguments
                dst = {'data_vars': dataset}
                for kw in ['data_vars', 'coords', 'attrs', 'dims']:
                    if kw in kwargs:
                        dst[kw] = kwargs[kw]
                self.dataset = xr.Dataset(**dst)
            else:
                # xarray from_dict
                self._create_from_dict(dataset)
        elif isinstance(dataset, Dataset):
            self.dataset = dataset.dataset
        else:
            raise TypeError(
                'Incorrect type {} to create a cerbere dataset object'
            )

        # add global attributes
        if attrs is not None:
            for attr, val in attrs.items():
                self.attrs[attr] = val

        if not isinstance(dataset, Dataset):
            if self.is_empty():
                return

        # standardize dataset
        self.dataset.cb.cerberize()
        # @TODO: not clear why the line above is not sufficient
        self.dataset = self.dataset.cb.cfdataset

        # set Z coordinate
        if as_z is not None:
            self.set_as_z(as_z)

    def set_as_z(self, fieldname: str, positive=None):
        """Set a field as representing the Z coordinate"""
        self.dataset.cb.set_as_z(fieldname, positive=positive)

    def to_xarray(self):
        """returns the Dataset as a xarray dataset object.

        Normally returns the underlying storage xarray dataset used in the
        Cerbere Dataset object if fitting the specifications of Cerbere well
        formed Dataset or Feature classes. If not, some transformation is
        performed to return a well formed xarray Dataset, which may take some
        computational time.
        """
        return self.dataset.cb.cfdataset

    @property
    def _dataset_class(self):
        return self.__class__.__name__

    def __str__(self):
        return self._std_dataset.cerbere.__str__()

    def _create_from_dict(self, data, **kwargs):
        """
        Create the dataset from a dict of fields and attributes.
        """
        # addition to xarray : fields can be provided as Field object
        def to_dict(arr):
            return {'dims': arr.dims, 'data': arr.data, 'attrs': arr.attrs}

        for var in data.keys():
            if isinstance(data[var], Field):
                data[var] = to_dict(data[var].to_xarray())

        if 'coords' in data.keys():
            for var, value in data['coords'].items():
                if isinstance(value, Field):
                    data[var] = to_dict(value.to_xarray())
        if 'data_vars' in data.keys():
            for var, value in data['data_vars'].items():
                if isinstance(value, Field):
                    data[var] = to_dict(value.to_xarray())

        # create a dataset
        self.dataset = xr.Dataset.from_dict(data)

        # assign coordinates
        for coord in REQUIRED_GEOCOORDINATES:
            if coord in self.dataset.data_vars:
                self.dataset = self.dataset.assign_coords(
                    {coord: self.dataset.data_vars[coord]}
                )

    def _check_mandatory_geocoordinates(self) -> bool:
        """check the required geolocation coordinates are in the dataset"""
        coord_validity = all([
            coord in self.coordnames
            for coord in REQUIRED_GEOCOORDINATES
        ])

        if not coord_validity:
            for coord in REQUIRED_GEOCOORDINATES:
                if coord not in self.coordnames:
                    logging.warning(
                        '  => Missing coordinate var: {}'.format(coord)
                    )
        return coord_validity

    def is_empty(self) -> bool:
        """Returns True if the dataset object contains no data yet"""
        return self.dataset.cb.cfdataset.equals(xr.Dataset())

    def __is_remote(self):
        """check is the url corresponds to a remote file (http, ftp,...)"""
        try:
            result = urlparse(self._url)
            return all([result.scheme, result.netloc])
        except ValueError:
            return False

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @classmethod
    def exists(cls, url: Union[str, Path]) -> bool:
        """Returns True if `url` is an existing resource"""
        if isinstance(url, Path):
            return url.exists()
        try:
            result = urlparse(url)
            if len(result.scheme) == 0:
                return os.path.exists(url)
            return all([result.scheme, result.netloc])
        except ValueError:
            return False

    @classmethod
    def _get_time_format(cls):
        return '%Y-%m-%dT%H:%M:%SZ'

    def has_field(self, fieldname: str) -> bool:
        """Return True if the field ``fieldname`` exists in the dataset."""
        return fieldname in self.fieldnames

    def rename_field(self, name: str, newname: str) -> None:
        """Rename a field.

        Args:
            name (str): name of the field to be renamed
            newname (str): new name of the field
        """
        self.dataset.cb.cfdataset = self.dataset.cb.cfdataset.rename(
            {name: newname})

    @property
    def url(self) -> Path:
        """Return the url of the file storing the dataset"""
        return self.dataset.cb.url

    @property
    def basename(self) -> str:
        """Return the basename of the file storing the dataset"""
        return os.path.basename(self.url)

    @property
    def _std_dataset(self) -> xr.Dataset:
        """The standardized xarray dataset object containing the data.

        Use with caution as dynamic transformations may not be applied (for
        better performances or memory occupation). You may retrieve a feature
        object that is no more generic. It is safer to use the
        :meth:`~cerbere.dataset.dataset.as_dataset` method (future release).
        """
        return self.dataset.cb.cfdataset

    @_std_dataset.setter
    def _std_dataset(self, value: Union['xr.Dataset', 'Dataset']) -> None:
        """The standardized version of the underlying xarray dataset object.

        Use with caution as dynamic transformations may not be applied (for
        better performances or memory occupation). You may retrieve a feature
        object that is no more generic. It is safer to use the
        :meth:`~cerbere.dataset.dataset.as_dataset` method (future release).
        """
        if not isinstance(value, (xr.Dataset, Dataset)):
            raise TypeError('Unknow type {} for _std_dataset'.format(
                type(value))
            )
        if isinstance(self.dataset, Dataset):
            self.dataset._std_dataset = value
        else:
            self.dataset = value

    def is_new(self) -> bool:
        """True if the storage is opened in (over)write mode"""
        return self._mode == OpenMode.WRITE_NEW

    def is_readonly(self) -> bool:
        """True if the file or resource is opened in read only mode"""
        return self._mode == OpenMode.READ_ONLY

    @property
    def filesize(self) -> int:
        """The dataset file size, in octets"""
        return self.dataset.cb.filesize()

    @property
    def file_creation_date(self) -> datetime.datetime:
        """The date the dataset file was generated"""
        return self.dataset.cb.file_creation_date()

    def has_coord(self, coord: Hashable) -> bool:
        """Return True if the coordinate field is defined"""
        return self.dataset.cb.has_coord(coord)

    @property
    def sizes(self) -> Mapping[Hashable, int]:
        """Mapping from dimension names to lengths."""
        return self.dims

    @property
    def dims(self) -> Mapping[Hashable, int]:
        """Mapping from dimension names to lengths.

        CF convention names are used whenever possible for the dimensions.
        """
        return OrderedDict(self._std_dataset.dims)

    @property
    def dimnames(self) -> Tuple[Hashable]:
        """Tuple of the dataset's dimension names"""
        return tuple(self.dims.keys())

    @property
    def dimsizes(self) -> Tuple[int]:
        """Tuple of the dataset's dimension sizes"""
        return tuple(self.dims.values())

    @property
    def attrs(self) -> MutableMapping[Hashable, Any]:
        """Mapping from global attribute names to value."""
        return self.dataset.attrs

    @attrs.setter
    def attrs(self, attrs):
        self.dataset.attrs = attrs

    def get_attr(self, attr: str) -> Any:
        """Returns the value of a global attribute.

        Args:
            attr(str): name of the global attribute.

        Returns:
            value of the requested attribute.
        """
        return self.dataset.attrs[attr]

    def get_dimsize(self, dimname: str) -> int:
        """Return the size of a dimension.

        Args:
            dimname (str): name of the dimension.

        Returns:
            int: size of the dimension.
        """
        return self.dims[dimname]

    def get_field_sizes(self, fieldname: Hashable) -> Mapping[Hashable, int]:
        """Mapping from dimension names to lengths of a field in the dataset

        Args:
            fieldname (str): name of the field
        """
        return OrderedDict(self._std_dataset[fieldname].sizes)

    def get_field_dims(self, fieldname: Hashable) -> Tuple[Hashable]:
        """Tuple of the dimension names of a field in the dataset

        Args:
            fieldname (str): name of the field
        """
        return tuple(self.get_field_sizes(fieldname).keys())

    @property
    def _varnames(self) -> List[Hashable]:
        """List of the names of all the fields (including coordinates) of the
        dataset.
        """
        return list(self._std_dataset.variables.keys())

    @property
    def coordnames(self) -> List[Hashable]:
        """List of names of the coordinate fields of the dataset."""
        return list(self._std_dataset.coords.keys())

    @property
    def geocoordnames(self) -> List[Hashable]:
        """List of names of the geospatial coordinate fields of the dataset."""
        return [
            k for k, v in self._std_dataset.coords.items() if 'axis' in v.attrs
        ]

    @property
    def coords(self) -> List['Field']:
        """List of coordinates (as Field objects)"""
        return [
            Field(v, name=k)
            for k, v in self._std_dataset.coords.items()
        ]

    @property
    def geocoords(self) -> List['Field']:
        """List of geolocation coordinates (as Field objects)"""
        return [
            Field(v, name=k)
            for k, v in self._std_dataset.coords.items()
            if 'axis' in v.attrs
        ]

    @property
    def fieldnames(self) -> List[Hashable]:
        """The names of the geophysical fields of the mapper.

        The coordinate field names are excluded from this list.

        Returns:
            list[str]: list of field names
        """
        return list(self._std_dataset.data_vars.keys())

    def get_field(self, fieldname: str) -> 'Field':
        """
        Return the :class:`~cerbere.dataset.field.Field` object corresponding to
        the requested field name.

        The :class:`~cerbere.dataset.field.Field` class contains all the metadata
        describing a field (equivalent to a variable in netCDF).

        Args:
            fieldname (str): name of the field

        Returns:
            the corresponding Field object
        """
        if fieldname not in self.fieldnames:
            raise ValueError('Unknown field {}'.format(fieldname))

        return Field(self._std_dataset[fieldname], fieldname, dataset=self)

    def add_field(self, field: 'Field', force_index: bool = True) -> None:
        """Add a field to the feature.

        Args:
            field: the field is provided as a
                :class:`~cerbere.dataset.field.Field` object
            force_index: if the added field contains an index coordinate with
                the same name as the dataset, replace the values with those of
                the dataset (otherwise only the field values for which the index
                values of the field and the dataset are the same will be added).
        """
        if field.name in self.fieldnames:
            raise Exception(
                'Field already existing in feature. Can not add {}'
                .format(field.name)
            )

        try:
            dataarr = field.to_xarray(silent=True, decoding=False)

            # if some indexes are existing in the dataset, ensure the values are
            # the same
            for idx in dataarr.indexes:
                if idx in self._std_dataset.indexes and force_index:
                    dataarr = dataarr.reset_index(idx, drop=True)

            self._std_dataset = self._std_dataset.assign(
                {field.name: dataarr}
            )
        except ValueError:
            # an error case when for instance an index (like time) has masked
            # values. This seems to do the trick, in some tested cases but
            # probably not all of them...
            warnings.warn(
                'Entering a special case where not xarray variable '
                'assignment was possible, probably because of fill or duplicate'
                ' values in a coordinate. Field: {}'.format(field.name))
            self._std_dataset[field.name] = xr.DataArray(
                coords=[(d, self._std_dataset.coords[d].data)
                        for d in field.dims],
                data=field.get_values(),
                attrs=field.attrs)
            self._std_dataset[field.name].encoding = field.encoding
        except:
            raise

    def drop_field(self, fieldname: str):
        """Remove a field from the feature"""
        self._std_dataset = self._std_dataset.drop(fieldname)

    def get_coord(self, coordname: str) -> 'Field':
        """Returns the coordinate field with the requested name.
        """
        if coordname not in self.coordnames:
            raise ValueError(
                'Coordinate field {} is not exising: '.format(coordname)
            )

        return Field(
            self._std_dataset.coords[coordname], name=coordname, dataset=self
        )

    def get_geocoord(self, coordname: str) -> 'Field':
        """Returns the geolocation coordinate field with the requested name.

        Possible coordinate field names are 'lat', 'lon', 'time', 'z'.
        """
        if coordname not in ['lat', 'lon', 'time', 'z']:
            logging.warning(
                '{} is not a geolocation coordinate'.format(coordname)
            )
        return self.get_coord(coordname)

    def get_geocoord_by_axis(self, axis: str) -> 'Field':
        """Returns the geolocation coordinate field for the given axis.

        Possible coordinate field names are 'lat', 'lon', 'time', 'z'.
        """
        if axis not in ['X', 'Y', 'T', 'Z']:
            logging.warning(
                '{} is not a geolocation coordinate'.format(axis)
            )
        for coordname in self.geocoordnames:
            if self.get_coord(coordname).attrs['axis'] == axis:
                return self.get_coord(coordname)

        raise ValueError('No coordinate found for axis: {}'.format(axis))

    @property
    def geodims(self) -> MutableMapping[str, int]:
        """Mapping from geolocation dimension names to lengths.
        """
        return OrderedDict([
            (DIM_MATCHING[_], self.sizes[_]) for _ in list(self.dimnames)
            if _ in DIM_MATCHING.keys()
            ])

    @property
    def geodimnames(self) -> Tuple[str]:
        """A tuple of the dataset geolocation dimension names"""
        return tuple(self.geodims.keys())

    @property
    def geodimsizes(self) -> Tuple[int]:
        """A tuple of the dataset geolocation dimension sizes"""
        return tuple(self.geodims.values())

    def get_field_fillvalue(self, fieldname: str) -> Any:
        """Returns the missing value of a field"""
        if fieldname in self.coordnames:
            return self.get_coord(fieldname).fill_value
        else:
            return self.get_field(fieldname).fill_value

    def get_values(self,
                   fieldname: str,
                   index: Mapping[str, slice] = None,
                   as_masked_array: bool = True,
                   expand: bool = False,
                   expand_dims: List[str] = None,
                   **kwargs):
        """Read the data of a field.

        Args:
            fieldname (str): name of the field which to read the data from

            index: any kind of xarray indexing compatible with
                :func:`xarray.DataArray.isel` selection method.

            padding: pad the result with fill values where slices are out of the
             field dimension limits. Default is False.

            as_masked_array (bool, optional): return the result as a masked
                array instead of a xarray DataArray. Default is True (may had some
                overhead).

        Return:
            MaskedArray: array of data read. Array type is the same as the
                storage type.
        """
        if expand_dims is not None:
            expand_dims = [self._std_dataset[_] for _ in expand_dims]

        return self._std_dataset[fieldname].cerbere.isel(
            index, as_masked_array=as_masked_array,
            broadcast_coords=expand_dims, **kwargs
        )

    def set_mask(self, fieldname: str, mask: np.ndarray):
        raise DeprecationWarning()

    def set_values(
            self,
            fieldname: str,
            values: Union[np.ndarray, np.ma.MaskedArray],
            index: Optional[Mapping[str, slice]] = None,
            **kwargs):
        """set the values of a field.

        It is  possible to set only a subset of the field data array, using
        ``index``:

        >>> import numpy as np
        >>> ...
        >>> dst.set_values(
        >>>        'test',
        >>>        np.full((10, 5,), 2),
        >>>        {'x': slice(10, 20), 'y': slice(0, 5)}
        >>>        )

        See:
        Field.set_values

        """
        self._std_dataset[fieldname].cerbere.loc[index] = values

    def clip_values(
            self,
            fieldname: Hashable,
            geometry: 'shapely.geometry.Polygon',
            masked: bool = False,
            continuity_threshold: int = 0,
            as_masked_array: bool = True,
            **kwargs) -> List[Union['xr.DataArray', 'np.ma.masked_array']]:
        """Return the data of a field within a geographical area.

        Extract the subset(s) of the Dataset including the whole area geometry,
        defined as a polygon of lon/lat coordinates.

        Any keywords from :meth:`get_values()` can be applied here too.

        Args:
            fieldname (str): name of the field which to read the data from

            geometry (shapely.geometry.Polygon): boundaries of the clipped data,
                as a polygon of lon, lat coordinates.

            continuity_threshold (int): maximum number of non-contiguous
                measurements allowed beyond which two datasets are extracted
                instead of one. Avoids creating too many subsets when lat/lon
                are not continuous.

            masked (bool): mask the data out of the geometry boundary.
                May take additional processing time.

            as_masked_array (bool): return result as a numpy masked array if
                True (default), or xarray DataArray if set to False.

        Return:
            data subset fitting the clip area, as a xarray DataArray or numpy
            masked array, depending on ``as_masked_array`` value.
        """
        (lonmin, latmin, lonmax, latmax) = geometry.bounds

        # get spatial dimensions
        spatial_dims = list(self.get_field_dims('lat'))
        for d in self.get_field_dims('lon'):
            if d not in spatial_dims:
                spatial_dims.append(d)

        # coarse search on geometry bounding box
        lat = self.get_values(
            'lat', expand=True, expand_dims=spatial_dims, as_masked_array=False,
            **kwargs
        )
        lon = self.get_values(
            'lon', expand=True, expand_dims=spatial_dims, as_masked_array=False,
            **kwargs
        )
        area_data = lat.where((
                (lon >= lonmin) & (lon <= lonmax) &
                (lat >= latmin) & (lat <= latmax)
        ))
        inside = ~area_data.to_masked_array().mask
        if continuity_threshold > 0:
            inside = binary_dilation(inside, iterations=continuity_threshold)
        slices = find_objects(label(inside)[0])

        # return all extracted subsets
        subsets = []
        for sl in slices:
            dslices = {d: sl[i] for i, d in enumerate(spatial_dims)}
            if 'index' in kwargs:
                if set(spatial_dims).intersection(kwargs['index'].keys()):
                    raise ValueError(
                        'slicing on spatial dimensions can not be used when '
                        'clipping data'
                    )
                dslices.update(kwargs.pop('index'))
            subsets.append(
                self.get_values(fieldname, index=dslices, **kwargs)
            )

            if masked:
                raise NotImplementedError

        return subsets

    def get_times(self, as_masked_array=True, **kwargs):
        """Return the times of a feature.

        The time values are returned as numbers. Use the
        :func:`get_time_units` function to convert these values to date or
        time objects.

        Subsets can be returned by using either slices or indices arguments.
        slices and indices are exclusive and can not be both specified. Both
        must be provided as dictionaries where keys are the names of the
        dimensions to subset. Values are slice objects in the case of `slice`
        argument and numbers in the case of `indices` argument. Only the
        subsetted dimensions need to be provided (the full range s assumed for
        the other dimensions).

        Args:
            index: any kind of xarray indexing
            as_masked_array (bool): return the result as a numpy masked array
                (by default), or as a numpy ndarray if set to False.
            padding (bool) : if True, pad the result with fill values where
                slices are out of the field size.

        Returns:
            numpy.ma.array: the requested data.
        """
        return self.dataset['time'].cerbere.isel(
            as_masked_array=True, **kwargs)

    def get_time_units(self):
        """Return the time units as a CF convention compliant string"""
        warnings.warn(
            'Deprecated get_time_units function.',
            FutureWarning
        )
        return DEFAULT_TIME_UNITS

    def get_datetimes(self, slices=None, indices=None, **kwargs):
        """Return the time values of a feature as datetime objects.

        Subsets can be returned by using either slices or indices arguments.
        slices and indices are exclusive and can not be both specified. Both
        must be provided as dictionaries where keys are the names of the
        dimensions to subset. Values are slice objects in the case of `slice`
        argument and numbers in the case of `indices` argument. Only the
        subsetted dimensions need to be provided (the full range s assumed for
        the other dimensions).

        Args:
            slices (dictionary): a dictionary where keys are the dimensions
                to slice and values are slice objects.

            indices (list): a dictionary where keys are the dimensions
                to slice and values are slice objects.

            cache (bool): if cache is True, the data read from file are kept in
                memory. The full field data array is kept in cache, slices or
                indices are ignored by caching (though the result returned by
                this function will be a subset if slices or indices are
                provided. Use the `view` concept in mappers instead if you want
                frequent accesses to a subset of a file.
                Default is False.

        Returns:
            numpy.ma.array<datetime>: the requested data.

        Warning:
            To be used with caution as conversion to datetime objects can take
            a long time. Use only when accessing single elements for instance.
        """

        if 'as_masked_array' not in kwargs or kwargs['as_masked_array']:
            return (self.get_values('time', **kwargs)
                    .astype('datetime64[us]')
                    .astype(datetime.datetime))
        else:
            raise NotImplementedError

    def get_lon(self, **kwargs):
        """Return the longitude values of a feature.

        Subsets can be returned by using either slices or indices arguments.
        slices and indices are exclusive and can not be both specified. Both
        must be provided as dictionaries where keys are the names of the
        dimensions to subset. Values are slice objects in the case of `slice`
        argument and numbers in the case of `indices` argument. Only the
        subsetted dimensions need to be provided (the full range s assumed for
        the other dimensions).

        Args:
            index: any kind of xarray indexing
            as_masked_array (bool): return the result as a numpy masked array
                (by default), or as a numpy ndarray if set to False.
            padding (bool) : if True, pad the result with fill values where
                slices are out of the field size.

        Returns:
            numpy.ma.array<float>: the requested longitudes.
        """
        return self.get_values('lon', **kwargs)

    def get_lat(self, **kwargs):
        """Return the latitude values of a feature.

        Subsets can be returned by using either slices or indices arguments.
        slices and indices are exclusive and can not be both specified. Both
        must be provided as dictionaries where keys are the names of the
        dimensions to subset. Values are slice objects in the case of `slice`
        argument and numbers in the case of `indices` argument. Only the
        subsetted dimensions need to be provided (the full range s assumed for
        the other dimensions).

        Args:
            index: any kind of xarray indexing
            as_masked_array (bool): return the result as a numpy masked array
                (by default), or as a numpy ndarray if set to False.
            padding (bool) : if True, pad the result with fill values where
                slices are out of the field size.

        Returns:
            numpy.ma.array<float>: the requested latitudes.
        """
        return self.get_values('lat', **kwargs)

    def set_z(self, values, ztype='depth'):
        """Set the depth(s)

        Args:
            values (:class:`numpy.ma.MaskedArray` or :class:`Field`): depth
                values, as a array of floats or a Field object

            ztype (string): type is depth (positive down) or height (positive
                up)
        """
        coord = self._make_coord(
            'z', values, 'm', ztype, ztype
        )
        coord.attrs.update(
            {'depth': {'positive': 'down'},
             'height': {'positive': 'up'}
             }[ztype]
        )
        return coord

    def get_z(self, **kwargs):
        """
        if cache is True, data read from storage are kept in memory
        """
        return self.get_values('z', **kwargs)

    def extract(
            self,
            index: Optional[Mapping[str, slice]] = None,
            fields: Optional[List[str]] = None,
            padding: bool = False,
            prefix: Optional[str] = None,
            deep: bool = True,
            **kwargs
    ) -> 'Dataset':
        """
        Extract a subset of the dataset.

        If ``deep`` is False, returns only a view on the dataset, sharing the
        same data arrays in memory as the subsetted dataset.

        Args:
            index: any kind of xarray indexing compatible with
                :func:`xarray.DataArray.isel` selection method.

            padding: pad the result with fill values where slices are out of the
             field dimension limits. Default is False. ``deep`` must be set to
             True.

            fields: list of field names to extract. If None, all fields
                are extracted.

            prefix: add a prefix string to the field names of the returned
                subset.
        """
        if fields is not None and len(fields) > 0:
            dst = self._std_dataset[fields].cerbere.isel(
                index, padding=padding, prefix=prefix, **kwargs)
        else:
            dst = self._std_dataset.cerbere.isel(
                index, padding=padding, prefix=prefix, **kwargs)

        if deep:
            dst = dst.copy(deep=True)

        return Dataset(dst)

    def clip(
            self,
            geometry: shapely.geometry.Polygon,
            masked: bool = False,
            continuity_threshold: int = 0,
            **kwargs) -> List['Dataset']:
        """Return the dataset subsets within a geographical area.

        Extract the subset(s) of the Dataset including the whole area geometry,
        defined as a polygon of lon/lat coordinates.

        Any keywords from :meth:`extract()` can be applied here too.

        Args:
            geometry (shapely.geometry.Polygon): boundaries of the clipped data,
                as a polygon of lon, lat coordinates.

            continuity_threshold (int): maximum number of non-contiguous
                measurements allowed beyond which two datasets are extracted
                instead of one. Avoids creating too many subsets when lat/lon
                are not continuous.

            masked (bool, optional): mask the data out of the geometry boundary.
                May take additional processing time.

        Return:
            MaskedArray: array of data read. Array type is the same as the
                storage type.
        """
        return [Dataset(_) for _ in self._std_dataset.cerbere.clip(
                    geometry=geometry, masked=masked,
                    continuity_threshold=continuity_threshold, **kwargs)]

    def get_location_by_value(self, field, value, **kwargs):
        """Return the indices and geolocation of a given value.

        If the value is not unique, the first one found is returned.
        """
        if not np.isscalar(value):
            raise ValueError('value must be a scalar')

        loc = np.where(self.get_values(field, **kwargs) == value)
        loc = {dim: loc[i][0]
               for i, dim in enumerate(self.get_field(field).dimnames)}

        geoloc = (np.asscalar(self.get_lon(index=loc)),
                  np.asscalar(self.get_lat(index=loc)))

        return loc, geoloc,

    def closest_spatial_location(self, lon, lat):
        """Get closest dataset lat/lon location to given coordinates.

        Use pythagorian differences on lat/lon values so take the result with
        caution.

        Returns:
            A tuple with the indices and lon/lat of the closest point found.
        """
        d_lat = self.get_lat() - lat
        d_lon = self.get_lon() - lon

        if len(d_lat.shape) == 2:
            r2 = d_lat ** 2 + d_lon ** 2
            loc = np.where(r2 == np.min(r2))
            loc = {dim: loc[i][0]
                   for i, dim in enumerate(self.get_field_dims('lat'))}
            geoloc = (np.asscalar(self.get_lon(index=loc)),
                      np.asscalar(self.get_lat(index=loc)))
            return loc, geoloc,

        else:
            # @TODO
            raise NotImplementedError

    def clone(
            self,
            fieldname: str,
            **kwargs
    ):
        """
        Create a copy of a field, limiting to a set of slices or indices, and
        padding out as required.

        See:
            :func:`cerbere.dataset.field.Field.clone`
        """
        return self.dataset[fieldname].isel(**kwargs).copy(deep=True)

    def get_field_attrs(self, fieldname) -> Mapping[Hashable, Any]:
        """Returns the attributes of a field.

        Args:
            fieldname (str): name of the field.

        Returns:
            a dictionary where keys are the attribute names.
        """
        if isinstance(self.dataset, Dataset):
            return self.dataset.get_field_attrs(fieldname)

        return self.dataset[fieldname].attrs

    @property
    def time_coverage_start(self) -> datetime.datetime:
        """Minimum sensing time of the data contained in the dataset"""
        return self._std_dataset.cerbere.time_coverage_start

    @time_coverage_start.setter
    def time_coverage_start(self, date: datetime.datetime):
        """Set the minimum sensing time of the data contained in the dataset
        """
        self._std_dataset.cerbere.time_coverage_start = date

    @property
    def time_coverage_end(self) -> datetime.datetime:
        """Maximum sensing time of the data contained in the dataset"""
        return self._std_dataset.cerbere.time_coverage_end

    @time_coverage_end.setter
    def time_coverage_end(self, date: datetime.datetime):
        """Set the maximum sensing time of the data contained in the dataset
        """
        self._std_dataset.cerbere.time_coverage_start = date

    @property
    def bbox(self) -> 'shapely.geometry.Polygon':
        """The bounding box, e.g. the south/north and east/west
        limits of the feature.
        """
        return self._std_dataset.cerbere.bbox

    @bbox.setter
    def bbox(self,
             bbox: Union[str, tuple, 'shapely.geometry.Polygon'] = 'auto'):
        """Set the bounding box of the feature

        Args:
            :class:`shapely.geometry.Polygon`: the bounding box. If None, it is
                estimated from the feature lat/lon.
        """
        self._std_dataset.cerbere.bbox = bbox

    @property
    def wkt_bbox(self) -> str:
        """The bounding box in WKT format."""
        return self._std_dataset.cerbere.bbox.wkt

    def _get_attr_value(self, att: str) -> Any:
        """Return global attribute value or None if the attribute does not
        exists.
        """
        return self._std_dataset.attrs.get(att)

    def get_product_version(self):
        """return the product version"""
        return self._std_dataset.cerbere.product_version

    def close(self):
        """Close file"""
        pass

    @staticmethod
    def _xr_fillvalue(xvar, default=None):
        """Return the fill value of an xarray

        Which can be either in attrs or encoding attributes
        """
        return xvar.encoding.get(
            '_FillValue', xvar.attrs.get('_FillValue', default))

    @staticmethod
    def _xr_units(xvar, default=None):
        """Return the units of an xarray

        Which can be either in attrs or encoding attributes
        """
        return xvar.encoding.get('units', xvar.attrs.get('units', default))

    def save(self,
             dest: Union[str, 'Dataset', None] = None,
             format: str = 'NETCDF4',
             profile: str = 'default_saving_profile.yaml',
             force_profile: bool = False,
             keep_src_encoding: bool = False):
        """
        Args:
            dest (str, optional): save to a new file, whose path is provided in
                this argument. Raise an exception if the file already exists.
            profile (str, optional): path to a specific format profile to
                apply before saving (or default formatting profile is used).
            force_profile (bool, optional): force profile attribute values to
                supersede existing ones in dataset attributes.
            keep_src_encoding (bool): keep original dtype, _FillValue
                and scaling if any (through `add_offset` or `scale_factor`
                attributes) as in the source data.
        """
        if not isinstance(dest, (str, Path)):
            raise DeprecationWarning

        return self.dataset.cb.save(
                path=dest,
                format=format,
                profile=profile,
                force_profile=force_profile
            )

        # save as a new file
        if dest is not None:
            if isinstance(dest, (str, Path)):
                # save in a file
                if os.path.exists(str(dest)):
                    raise IOError('This file already exists')

                self._url = str(dest)
                self._mode = OpenMode.WRITE_NEW
                self._format = format
            elif isinstance(dest, Dataset):
                # if not dest.is_empty():
                #     raise IOError(
                #         'Can not save into a dataset that is not empty.'
                #     )
                if dest.is_readonly():
                    raise IOError('destination dataset is read-only')
                # use dest dataset as formatter
                dest.dataset = self.dataset.copy()
                return dest.save(
                    format=format,
                    profile=profile,
                    force_profile=force_profile
                )

        # save in the file currently attached to the dataset
        else:
            if self._mode == OpenMode.READ_ONLY:
                raise IOError(
                    'Can not save a dataset open in read only mode or with no '
                    'url provided'
                )

        # apply new formatting rules
        saved_dataset = self._convert_format(
            profile=profile, force_profile=force_profile)

        # remove internal special attributes
        for v in saved_dataset.variables.values():
            for att in ['_attached_dataset', '_components']:
                if att in v.attrs:
                    v.attrs.pop(att)

        # save to chosen format
        if 'NETCDF' in self._format:
            self._to_netcdf(saved_dataset, keep_src_encoding)

        else:
            logging.error('Unknown output format : {}'.format(self._format))
            raise NotImplementedError

    def add_global_attrs(
            cls,
            dataset: 'xr.Dataset',
            attrs: Mapping[str, Any],
            force_profile: bool = False) -> None:
        """Add default attributes to the dataset from a attribute definition
        file.

        Args:
            attrs (dict): the global attributes to add when saving the dataset
        """
        # add attributes
        # don't override previous attribute values
        for att in attrs:
            if att not in dataset.attrs or force_profile:
                if attrs[att] is None:
                    continue
                dataset.attrs[att] = attrs[att]

    def add_field_attrs(
            cls,
            dataset: xr.Dataset,
            attrs: Mapping[str, Any],
            force_profile: bool = False) -> None:
        """Add field attributes to the dataset from a attribute definition
        file.

        Args:
            attrs (dict): the field attributes to add when saving the dataset
        """
        # add attributes
        # don't override previous attribute values
        for v in attrs:
            if v not in dataset.variables:
                logging.warning(
                    'Field {} not found in the dataset to save. Skipping'
                    .format(v)
                )
                continue
            for att in attrs[v]:
                if att not in dataset.variables[v].attrs or force_profile:
                    value = attrs[v][att]
                    if value is None:
                        continue
                    dataset.variables[v].attrs[att] = value

    def add_field_encoding(
            cls,
            dataset: 'xr.Dataset',
            attrs: Mapping[str, Any],
            force_profile: bool = False) -> None:
        """Add field encoding to the dataset from a attribute definition
        file.

        Args:
            attrs (dict): the field encoding attributes to add when saving the
                dataset
        """
        # add attributes
        # overrides previous attribute values!!
        for v in attrs:
            if v not in dataset.variables:
                logging.warning(
                    'Field {} not found in the dataset to save. Skipping'
                    .format(v)
                )
                continue
            for att in attrs[v]:
                if att not in dataset.variables[v].encoding or force_profile:
                    value = attrs[v][att]
                    if att == 'dtype':
                        att = C_DTYPE
                    if value is None:
                        continue
                    dataset.variables[v].encoding[att] = value

    @property
    def collection_id(self) -> str:
        """return the identifier of the product collection"""
        return self.dataset.cb.collection_id

    def get_naming_authority(self) -> str:
        return self.dataset.cb.naming_authority



from __future__ import annotations


from io import BufferedIOBase
from importlib_metadata import entry_points
import logging
import os
from pathlib import Path
import typing as T
import warnings

import numpy as np
import xarray as xr
from xarray.backends.common import AbstractDataStore

import cerbere.accessor.cdataarray
import cerbere.accessor.cdataset
import cerbere.cfconvention as cf
from .reader.basereader import BaseReader
from .feature.cbasefeature import BaseFeature
from .feature.cdiscretefeature import DiscreteFeature
from .feature.cbasecollection import BaseCollection


_datasets = {
        entry_point.name: entry_point
        for entry_point
        in entry_points().select(group='cerbere.plugins')
}

_readers = {
        entry_point.name: entry_point
        for entry_point
        in entry_points().select(group='cerbere.reader')
}

_features = {
        entry_point.name: entry_point
        for entry_point
        in entry_points().select(group='cerbere.features')
}

_feature = {
        entry_point.name: entry_point
        for entry_point
        in entry_points().select(group='cerbere.feature')
}



# Preserve the data dtype during masked array to xarray conversions (whe
# creating a DataArray from a numpy MaskedArray or from data with a
# _FillValue encoding attribute
_keep_dtype: bool = False


def set_options(keep_dtype: bool = True,
                longitude180: bool = True) -> None:
    """
    """
    global _keep_dtype, _longitude180
    _keep_dtype = keep_dtype
    _longitude180 = longitude180


def datasets():
    warnings.warn(
        'Usage of plugins is now deprecated and is replaced with xarray '
        'backend classes.', DeprecationWarning, stacklevel=2)
    return list(_datasets.keys())


def readers():
    return list(_readers.keys())


def _open_dataset(
        url: T.Union[str, Path],
        reader: BaseReader,
        **kwargs) -> xr.Dataset:
    """
    Open a file (netCDF, ZArr,...) and returns its content as a xarray_
    :class:`xarray.Dataset` object.
    """
    try:
        options = kwargs
        # for opt in ['mask_and_scale',
        #             'decode_cf',
        #             'decode_coords',
        #             'decode_times',
        #             'cache'
        #             ]:
        #     if opt not in options:
        #         options[opt] = True

        if isinstance(url, list):
            if 'combine' not in options:
                options['combine'] = 'by_coords'
            return reader.open_mfdataset(url, **options)
        else:
            return reader.open_dataset(url, **options)
    except IOError:
        logging.error('Could not read file: {}'.format(url))
        raise


def guess_reader_class(path) -> T.Type[BaseReader]:
    """Use the file patterns defined for each reader class to infer which
    reader class to use."""
    filename = Path(path).name

    candidates = []
    for rd in readers():
        # default reader : use only if no reader overrides this one
        if rd == 'xarray':
            continue
        tested_reader = _readers[rd].load()
        if tested_reader.guess_can_open(filename):
            candidates.append(tested_reader)

    if len(candidates) > 1:
        logging.warning(f'multiple readers matching this file pattern were '
                        f'found: {candidates}. Using the first of this list.')

    if len(candidates) > 0:
        return candidates[0]

    # defaults to xarray readers
    return BaseReader


def guess_feature(dataset: xr.Dataset) -> cerbere.feature.cfeature.BaseFeature:
    candidates = []
    for feat in _feature:
        feature_class = _feature[feat].load()
        match = feature_class.guess_feature(dataset)
        if match is not None:
            candidates.append(match)

    if len(candidates) == 1:
        return candidates[0]

    if len(candidates) > 1:
        # group aliases
        aliases = {
            _: _.is_alias(dataset) for _ in candidates
            if _.is_alias(dataset) is not None}
        if len(aliases) == 1:
            if set(candidates) == \
                    set(aliases.keys()).union(set(aliases.values())):
                return list(aliases.keys())[0]
        logging.warning(
            f'Several features matching the dataset: {candidates}. Cerbere '
            f'can not decide unambiguously which one should be used and will '
            f'use: {candidates[0]}')
        return candidates[0]

    # raise Exception if no feature could be matched
    raise TypeError(
        'No feature class could be found matching this dataset. There may be '
        'to many non compliance to CF in this dataset and a reader class '
        'should be created to fix this.')


def as_feature(dataset: xr.Dataset) -> cerbere.feature.cfeature.BaseFeature:
    return guess_feature(dataset)(dataset)


def open_dataset(
        dataset: T.Union[str, Path],
        reader: T.Optional[str] = None,
        plugin: T.Optional[str] = None,
        **kwargs
        ):
    """
    Open a cerbere Dataset object.

    The following xarray options are always set to True by cerbere when
    opening a file:
    * mask_and_scale
    * decode_cf
    * decode_coords
    * decode_coords

    Args:
        dataset:  similar to ``dataset`` arg in Dataset base class
        plugin: Dataset class name
    """
    for option in ['mask_and_scale', 'decode_cf', 'decode_coords',
                   'decode_coords']:
        if option in kwargs:
            warnings.WarningMessage(
                f'option {option} is always set to True by cerbere. Your '
                f'argument setting for this option will be ignored.')

    # DEPRECATED - for compatibility with cerbere v2
    if plugin is not None:
        warnings.warn(
            'Usage of plugins is now deprecated and is replaced with xarray '
            'backend classes.', DeprecationWarning, stacklevel=2)
        try:
            # for compatibility with cerbere v2
            old_dataset_class = _datasets[plugin].load()
            engine = old_dataset_class.engine
        except KeyError:
            raise TypeError("Can't find plugin for: {}. Did you install it?"
                            .format(plugin))
        obj = _open_dataset(dataset, engine=engine, **kwargs)
        if obj is None:
            raise IOError('Dataset not correctly opened')

        # cerbere sets
        obj.cb.url = dataset

        return obj

    # Cerbere v3
    if reader is None:
        # try to guess which reader to use
        reader_class = guess_reader_class(dataset)
    else:
        reader_class = _readers[reader].load()
    if reader_class is not None:
        logging.debug(f'Reader class for this file: {reader_class}')

    obj = _open_dataset(
        dataset, reader=reader_class, **kwargs).pipe(
        reader_class.postprocess)
    if obj is None:
        raise IOError('Dataset not correctly opened')

    # cerbere sets
    obj.cb.url = dataset

    return obj.cb.cfdataset


def open_feature(
        filename_or_obj: str | os.PathLike[T.Any] | BufferedIOBase |
                         AbstractDataStore | xr.Dataset,
        feature: str = None,
        reader: str = None,
        **kwargs
        ) -> BaseFeature:
    """
    Open a cerbere Dataset into a feature object.

    Args:
        filename_or_obj: path to the file, as in Xarray ``open_dataset``
        feature: Feature class name. If None, cerbere will attempt to guess
            the feature class and return an object of this class
        reader: reader class name
    """
    if not isinstance(filename_or_obj, xr.Dataset):
        dst = cerbere.open_dataset(filename_or_obj, reader=reader, **kwargs)
    else:
        dst = filename_or_obj

    if feature is not None:
        return _feature[feature].load()(dst)

    # try to guess the feature
    guessed_feature = guess_feature(dst)
    if (issubclass(guessed_feature, BaseCollection) and
            guessed_feature.is_alias(dst) is None):
        instance_class = guessed_feature.infer_instance_class(dst)
        return guessed_feature(dst, instance_class=instance_class)
    return guessed_feature(dst)


def open_as_feature(
        feature: str,
        dataset: T.Union[str, Path] = None,
        plugin: str = 'Dataset',
        **kwargs
        ):
    """
    Open a cerbere Dataset into a feature object.

    Args:
        feature: Feature class name
        dataset:  similar to ``dataset`` arg in Dataset base class
        plugin: Dataset class name
    """
    warnings.warn(
        'Usage of Dataset classes is now deprecated and will be removed '
        'in cerbere 3.1+', DeprecationWarning, stacklevel=2)
    dataset = _datasets[plugin].load()(dataset, **kwargs)

    return _features[feature].load()(dataset)


def new_array(
        data: T.Any,
        name: T.Optional[T.Hashable] = None,
        science_dtype: T.Optional[T.Union[np.dtype, T.Type[float]]] = None,
        fill_value: T.Optional[T.Any] = None,
        precision: T.Optional[int] = None,
        long_name: T.Optional[str] = None,
        standard_name: T.Optional[T.Union[str, T.Tuple[str, str]]] = None,
        units: T.Optional[str] = None,
        ancillary_variables: T.Optional[T.List[str]] = None,
        set_longitude180: bool = True,
        **kwargs) -> xr.DataArray:
    """
    Creates a DataArray with cerbere properties.

    Creating a DataArray, even if not using the cerbere properties,
    helps to solve some shortcomings of xarray. In particular:
    * preserving the fill_value associated with MaskedArray
    * preserving the original science dtype (as xarray may change the
      array dtype during some operations.

    Args:
        dtype: the type of the data. Infer the type from the provided data by
            default. xarray changes the data type of arrays when using masking,
            defining the field true type with this keyword will ensure you
            always get the data in the expected dtype when accessing with
            `get_values` method. Data will also be saved with the expected
            dtype.

        precision (int, optional): number of significant digits (when writing
            the data on file)

        set_longitude180 (bool): set the longitude to be between -180/180 (
            longitudes between 0-360 will be modified when reading the
            coordinate)

        fields : the subfields composing the main field.
            This is intended to group for instance components of the same
            variable (such as a vector's northward and eastward components
            for wind, currents,...). This allows to relate these components
            to the same physical variable (e.g wind) and to a single
            qc_levels and qc_details information.

        fill_value : the default value to associate with missing data in the
            field's values. The fill value must be of the same type as
            `dtype`. If None, default fill value will be used
            unless `no_missing_value` is set.

        no_missing_value: if True, the field will not contain any missing value.
            In practice no _FillValue attribute will be written for this field
            when saving it. Use this for mask or flag fields for instance.

        units : the units in which the data are given (if
            applicable)

        description (str) : full name of the phenomenon. This corresponds
            to a long_name in attribute in a netCDF file.

        authority (str): naming authority referencing the provided
            standard name

        standard_name (optional, str): standard label for a phenomenon, with
            respect to the convention stated in `authority` argument.
            This corresponds to a standard_name attribute in a CF compliant
            NetCDF file.

        quality_vars: list of related quality fields
    """
    global _keep_dtype

    dtype = data.dtype

    if isinstance(data, xr.DataArray):
        array = data

    else:
        if isinstance(data, np.ma.MaskedArray):
            if fill_value is None:
                fill_value = data.dtype.type(data.fill_value)

            if _keep_dtype:
                data = data.filled(fill_value)

        array = xr.DataArray(data, **kwargs)

    if dtype != array.dtype:
        if _keep_dtype:
            raise ValueError("xarray tried to change the data dtype despite "
                             "the keep_dtype cerbere option is set.")

        # the data dtype was changed which is because the DataArray is
        # initialized from a numpy MaskedArray with actual masked data.
        # We need to ensure to keep all the information to reverse the
        # dtype conversion
        if not isinstance(data, np.ma.MaskedArray):
            raise TypeError('A MaskedArray is expected here... Check the '
                            'cerbere code, this must be a bug.')
        # ensure the information on the original dtype is preserved and there
        # is a fill_value defined
        if fill_value is None:
            # use same fill value as xarray would use
            fill_value = cf.default_fill_value(dtype)
        array.cb._cfarray.encoding['_FillValue'] = fill_value

    # standardize according to Cerbere CF and ACDD conventions
    array.cb.cerberize()

    if name is not None:
        array.name = name

    # ensure the information on the original dtype is preserved
    if science_dtype is not None:
        array.cb.science_dtype = science_dtype
    else:
        array.cb.science_dtype = dtype

    # fill value
    if fill_value is not None:
        array.cb.fill_value = fill_value

    if standard_name is not None:
        array.cb.standard_name = standard_name
    if long_name is not None:
        array.cb.description = long_name
    if precision is not None:
        array.cb.precision = precision
    if units is not None:
        array.cb.units = units
    if ancillary_variables is not None:
        array.cb.quality_vars = ancillary_variables

    if set_longitude180:
        array.cb.longitude180 = set_longitude180

    return array


def reader_class(name: str):
    return _readers[name].load()


def feature_class(name: str):
    return _feature[name].load()


def class_by_cf_feature_type(name: str) -> BaseFeature:
    """Return the feature class corresponding to a CF featureType attribute
    value"""
    for feat in _feature:
        feature_type = _feature[feat].load()
        if (issubclass(feature_type, DiscreteFeature) and
                feature_type.cf_feature_type is not None and
                feature_type.cf_feature_type.lower() == name.lower()):
            return feature_type



# encoding: utf-8
"""
Abstract class for discrete data feature objects
"""
import typing as T

import xarray as xr

from .cbasefeature import BaseFeature


class DiscreteFeature(BaseFeature):

    # Each type of discrete sampling geometry (point, time series, profile or
    # trajectory) is defined by the relationships among its spatiotemporal
    # coordinates. We refer to the type of discrete sampling geometry as its
    # featureType. The term “ feature ” refers herein to a single instance of
    # the discrete sampling geometry (such as a single time series).
    # the following class attributes define the expected dimensions and shape
    # of the CF features

    # must be defined for any feature corresponding to a CF discrete sampling
    # geometry, Nobe otherwise
    cf_feature_type = None

    # the instance dimension name for collection of features
    cf_instance_axis = None

    # the instance dimension name for collection of features which are
    # themselves collections (like TimeSeriesProfile, TrajectoryProfile,...)
    cf_element_instance_axis = None

    @classmethod
    def _unexpected_featureType(cls, cfdst):
        """check CF featureType"""
        if cls.cf_feature_type is None or cfdst.cb.featureType is None:
            return False
        return cfdst.cb.featureType.lower() != cls.cf_feature_type.lower()

    @classmethod
    def _canonical_dims(
            cls, coord: str, *args, **kwargs
    ) -> T.Union[T.Set[str], T.List[T.Set[str]]]:
        if isinstance(cls.cf_canonical_dims[coord], list):
            return [set(_) - {cls.cf_instance_axis}
                    for _ in cls.cf_canonical_dims[coord]]
        return set(cls.cf_canonical_dims[coord]) - {cls.cf_instance_axis}

    @classmethod
    def cerberize(
            cls,
            dataset: xr.Dataset,
            force_reorder: bool = True,
            instance_class: 'BaseFeature' = None
    ) -> xr.Dataset:

        ds = super().cerberize(
            dataset, force_reorder=force_reorder, instance_class=instance_class)

        # fill in feature type attributes
        if cls.cf_feature_type is not None:
            ds.attrs['featureType'] = cls.cf_feature_type

        return ds



# -*- coding: utf-8 -*-
"""
Class for the image feature.
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

from cerbere.feature.feature import Feature

__all__ = ['Image']


class Image(Feature):
    """
    Feature class for representing a image, a two-dimensional irregular grid
    along the satellite track, with one single time value associated.

    An image feature typically corresponds to a satellite "snapshot" like
    provided typically with a SAR (though technically it is more a swath) or a
    high resolution optical imager: a spatially limited swath section, so small
    that we don't need a time value for each pixel or even scan line. A single
    time value is enough to describe it, as if it was an instant snapshot.
    """
    _feature_geodimnames = 'row', 'cell',

    def __init__(self, *args, **kwargs):
        # create feature
        super(Image, self).__init__(
            *args,
            **kwargs
        )

        # squeeze time dimension in non-coordinate fields
        ctime = self._std_dataset.time.copy()
        if 'time' in self._std_dataset.dims:
            self._std_dataset = self._std_dataset.squeeze(dim='time')
        self._std_dataset.coords['time'] = ctime

    def get_geocoord_dimnames(
            self, fieldname: str,
            values: 'xr.DataArray') -> Tuple[str, ...]:
        if fieldname in ['depth', 'height', 'alt']:
            return 'z',
        elif fieldname == 'time':
            return 'time',
        else:
            return 'row', 'cell',
